import cv2, time, math, contextlib

import numpy as np
import torch, torchvision
import torch.nn.functional as F

from .nms import nms


def empty_like(x):
    """
    Create empty torch.Tensor or np.ndarray with same shape as input and float32 dtype.
    
    Function:
        Create empty container with same shape as input for improved memory allocation efficiency.
    
    Input:
        x (torch.Tensor | np.ndarray): Input tensor or array
    
    Output:
        torch.Tensor | np.ndarray: Empty tensor or array with same shape as input, dtype float32
    
    Example:
        >>> tensor = torch.rand(2, 3, 4)
        >>> empty_tensor = empty_like(tensor)
        >>> print(empty_tensor.shape)  # torch.Size([2, 3, 4])
    """
    return (
        torch.empty_like(x, dtype=torch.float32) if isinstance(x, torch.Tensor) else np.empty_like(x, dtype=np.float32)
    )


def xywh2xyxy(x):
    """
    Convert bounding box coordinates from (x, y, width, height) format to (x1, y1, x2, y2) format.
    
    Function:
        Convert center point coordinates and width/height format to top-left and bottom-right format
        where (x1, y1) is the top-left corner and (x2, y2) is the bottom-right corner.
    
    Input:
        x (np.ndarray | torch.Tensor): Input bounding box coordinates in (x, y, width, height) format
                                      with shape (..., 4)
    
    Output:
        np.ndarray | torch.Tensor: Converted bounding box coordinates in (x1, y1, x2, y2) format
                                  with same shape as input
    
    Example:
        >>> boxes = torch.tensor([[100, 100, 50, 30]])  # center(100,100), width=50, height=30
        >>> xyxy_boxes = xywh2xyxy(boxes)
        >>> print(xyxy_boxes)  # [[75, 85, 125, 115]]
    """
    assert x.shape[-1] == 4, f"input shape last dimension expected 4 but input shape is {x.shape}"
    y = empty_like(x)  # faster than clone/copy
    xy = x[..., :2]  # centers
    wh = x[..., 2:] / 2  # half width-height
    y[..., :2] = xy - wh  # top left xy
    y[..., 2:] = xy + wh  # bottom right xy
    return y


def xywh2xyxy(x):
    """
    Convert bounding box coordinates from (x, y, width, height) format to (x1, y1, x2, y2) format.
    
    Function:
        Convert center point coordinates and width/height format to top-left and bottom-right format
        where (x1, y1) is the top-left corner and (x2, y2) is the bottom-right corner.
    
    Input:
        x (np.ndarray | torch.Tensor): Input bounding box coordinates in (x, y, width, height) format
                                      with shape (..., 4)
    
    Output:
        np.ndarray | torch.Tensor: Converted bounding box coordinates in (x1, y1, x2, y2) format
                                  with same shape as input
    
    Example:
        >>> boxes = torch.tensor([[100, 100, 50, 30]])  # center(100,100), width=50, height=30
        >>> xyxy_boxes = xywh2xyxy(boxes)
        >>> print(xyxy_boxes)  # [[75, 85, 125, 115]]
    """
    assert x.shape[-1] == 4, f"input shape last dimension expected 4 but input shape is {x.shape}"
    y = empty_like(x)  # faster than clone/copy
    xy = x[..., :2]  # centers
    wh = x[..., 2:] / 2  # half width-height
    y[..., :2] = xy - wh  # top left xy
    y[..., 2:] = xy + wh  # bottom right xy
    return y

def xyxy2xywh(x):
    """
    Convert bounding box coordinates from (x1, y1, x2, y2) format to (x, y, width, height) format where (x1, y1) is the
    top-left corner and (x2, y2) is the bottom-right corner.

    Args:
        x (np.ndarray | torch.Tensor): Input bounding box coordinates in (x1, y1, x2, y2) format.

    Returns:
        (np.ndarray | torch.Tensor): Bounding box coordinates in (x, y, width, height) format.
    """
    assert x.shape[-1] == 4, f"input shape last dimension expected 4 but input shape is {x.shape}"
    y = empty_like(x)  # faster than clone/copy
    y[..., 0] = (x[..., 0] + x[..., 2]) / 2  # x center
    y[..., 1] = (x[..., 1] + x[..., 3]) / 2  # y center
    y[..., 2] = x[..., 2] - x[..., 0]  # width
    y[..., 3] = x[..., 3] - x[..., 1]  # height
    return y
    

def xywhr2xyxyxyxy(x):
    """
    Convert batched Oriented Bounding Boxes (OBB) from [xywh, rotation] to [xy1, xy2, xy3, xy4] format.

    Args:
        x (np.ndarray | torch.Tensor): Boxes in [cx, cy, w, h, rotation] format with shape (N, 5) or (B, N, 5).
            Rotation values should be in radians from 0 to pi/2.

    Returns:
        (np.ndarray | torch.Tensor): Converted corner points with shape (N, 4, 2) or (B, N, 4, 2).
    """
    cos, sin, cat, stack = (
        (torch.cos, torch.sin, torch.cat, torch.stack)
        if isinstance(x, torch.Tensor)
        else (np.cos, np.sin, np.concatenate, np.stack)
    )

    ctr = x[..., :2]
    w, h, angle = (x[..., i : i + 1] for i in range(2, 5))
    cos_value, sin_value = cos(angle), sin(angle)
    vec1 = [w / 2 * cos_value, w / 2 * sin_value]
    vec2 = [-h / 2 * sin_value, h / 2 * cos_value]
    vec1 = cat(vec1, -1)
    vec2 = cat(vec2, -1)
    pt1 = ctr + vec1 + vec2
    pt2 = ctr + vec1 - vec2
    pt3 = ctr - vec1 - vec2
    pt4 = ctr - vec1 + vec2
    return stack([pt1, pt2, pt3, pt4], -2)


def clip_boxes(boxes, shape):
    """
    Clip bounding boxes to image boundaries to prevent exceeding image bounds.
    
    Function:
        Restrict bounding box coordinates within image bounds [0, width] and [0, height]
        to ensure boxes don't exceed image boundaries.
    
    Input:
        boxes (torch.Tensor | numpy.ndarray): Bounding boxes to clip in (x1, y1, x2, y2) format
                                              with shape (..., 4)
        shape (tuple): Image shape (height, width)
    
    Output:
        numpy.ndarray: Clipped bounding boxes with coordinates restricted to image bounds
                      with same shape as input
    
    Example:
        >>> boxes = torch.tensor([[-10, -5, 650, 480]])  # boxes exceeding boundaries
        >>> shape = (480, 640)  # image dimensions
        >>> clipped = clip_boxes(boxes, shape)
        >>> print(clipped)  # [[0, 0, 640, 480]]
    """
    boxes[..., [0, 2]] = boxes[..., [0, 2]].clip(0, shape[1])  # x1, x2
    boxes[..., [1, 3]] = boxes[..., [1, 3]].clip(0, shape[0])  # y1, y2

    return boxes


def scale_boxes(img1_shape, boxes, img0_shape, ratio_pad=None, padding=True, xywh=False):
    """
    Scale bounding boxes from one image size to another image size.
    
    Function:
        Scale bounding box coordinates from processed image size back to original image size.
        Supports YOLO-style padding and different bounding box formats.
    
    Input:
        img1_shape (tuple): Image shape corresponding to bounding boxes (height, width)
        boxes (torch.Tensor): Bounding box coordinates in (x1, y1, x2, y2) format
        img0_shape (tuple): Target image shape (height, width)
        ratio_pad (tuple, optional): Scale ratio and padding (ratio, pad), auto-calculated if not provided
        padding (bool): Whether to consider YOLO-style padding, default True
        xywh (bool): Whether bounding box format is xywh, default False
    
    Output:
        numpy.ndarray: Scaled bounding boxes in (x1, y1, x2, y2) format
    
    Example:
        >>> img1_shape = (640, 640)  # processed image size
        >>> boxes = torch.tensor([[100, 100, 200, 200]])
        >>> img0_shape = (480, 640)  # original image size
        >>> scaled_boxes = scale_boxes(img1_shape, boxes, img0_shape)
    """
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (
            round((img1_shape[1] - img0_shape[1] * gain) / 2 - 0.1),
            round((img1_shape[0] - img0_shape[0] * gain) / 2 - 0.1),
        )  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    if padding:
        boxes[..., 0] -= pad[0]  # x padding
        boxes[..., 1] -= pad[1]  # y padding
        if not xywh:
            boxes[..., 2] -= pad[0]  # x padding
            boxes[..., 3] -= pad[1]  # y padding
    boxes[..., :4] /= gain
    return clip_boxes(boxes, img0_shape)


def scale_coords(img1_shape, coords, img0_shape, ratio_pad=None, normalize: bool = False, padding: bool = True):
    """
    Scale segmentation coordinates from one image size to another image size.
    
    Function:
        Scale segmentation coordinates from processed image size back to original image size.
        Supports coordinate normalization and YOLO-style padding handling.
    
    Input:
        img1_shape (tuple): Source image shape (height, width)
        coords (torch.Tensor): Coordinates to scale with shape (N, 2)
        img0_shape (tuple): Target image shape (height, width)
        ratio_pad (tuple, optional): Scale ratio and padding values ((ratio_h, ratio_w), (pad_h, pad_w))
        normalize (bool): Whether to normalize coordinates to range [0, 1], default False
        padding (bool): Whether coordinates are based on YOLO-style padded images, default True
    
    Output:
        torch.Tensor: Scaled coordinates
    
    Example:
        >>> img1_shape = (640, 640)  # processed image size
        >>> coords = torch.tensor([[100, 100], [200, 200]])
        >>> img0_shape = (480, 640)  # original image size
        >>> scaled_coords = scale_coords(img1_shape, coords, img0_shape)
    """
    if coords.shape[0] == 0:
        return torch.zeros((17, 3), device=coords.device)
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0]
        pad = ratio_pad[1]

    if padding:
        coords[..., 0] -= pad[0]  # x padding
        coords[..., 1] -= pad[1]  # y padding
    coords[..., 0] /= gain
    coords[..., 1] /= gain
    coords = clip_coords(coords, img0_shape)
    if normalize:
        coords[..., 0] /= img0_shape[1]  # width
        coords[..., 1] /= img0_shape[0]  # height
    return coords


def clip_coords(coords, shape):
    """
    Clip line coordinates to the image boundaries.
    
    Function:
        Restrict coordinates within image bounds [0, width] and [0, height]
        to prevent coordinates from exceeding image boundaries.
    
    Input:
        coords (torch.Tensor | numpy.ndarray): Line coordinates list with shape (..., 2)
        shape (tuple): Image dimensions (height, width)
    
    Output:
        torch.Tensor | numpy.ndarray: Clipped coordinates with same type as input
    
    Example:
        >>> coords = torch.tensor([[-10, -5], [650, 480]])  # coordinates exceeding boundaries
        >>> shape = (480, 640)  # image dimensions
        >>> clipped = clip_coords(coords, shape)
        >>> print(clipped)  # [[0, 0], [640, 480]]
    """
    if isinstance(coords, torch.Tensor):  # faster individually (WARNING: inplace .clamp_() Apple MPS bug)
        coords[..., 0] = coords[..., 0].clamp(0, shape[1])  # x
        coords[..., 1] = coords[..., 1].clamp(0, shape[0])  # y
    else:  # np.array (faster grouped)
        coords[..., 0] = coords[..., 0].clip(0, shape[1])  # x
        coords[..., 1] = coords[..., 1].clip(0, shape[0])  # y
    return coords


def scale_coordsV5(img1_shape, coords, img0_shape, ratio_pad=None):
    """
    YOLOv5 version coordinate scaling function.
    
    Function:
        Specialized coordinate scaling implementation for YOLOv5.
        Scale coordinates from processed image size back to original image size.
    
    Input:
        img1_shape (tuple): Source image shape (height, width)
        coords (torch.Tensor): Coordinates to scale with shape (N, 2)
        img0_shape (tuple): Target image shape (height, width)
        ratio_pad (tuple, optional): Scale ratio and padding values, auto-calculated if not provided
    
    Output:
        torch.Tensor: Scaled coordinates, rounded to nearest integer
    
    Example:
        >>> img1_shape = (640, 640)  # processed image size
        >>> coords = torch.tensor([[100, 100], [200, 200]])
        >>> img0_shape = (480, 640)  # original image size
        >>> scaled_coords = scale_coordsV5(img1_shape, coords, img0_shape)
    """
    def clip_coords(boxes, img_shape, step=2):
        # Clip bounding xyxy bounding boxes to image shape (height, width)
        boxes[:, 0::step].clamp_(0, img_shape[1])  # x1
        boxes[:, 1::step].clamp_(0, img_shape[0])  # y1
        
    step=1
    # Rescale coords (xyxy) from img1_shape to img0_shape
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0]
        pad = ratio_pad[1]
    if isinstance(gain, (list, tuple)):
        gain = gain[0]
    coords[:, 0::step] -= pad[0]  # x padding
    coords[:, 1::step] -= pad[1]  # y padding
    coords[:, 0::step] /= gain
    coords[:, 1::step] /= gain
    clip_coords(coords, img0_shape, step=step)
    coords = coords.round()
    return coords


def box_area(box):
    """
    Calculate the area of bounding boxes.
    
    Function:
        Calculate area based on top-left and bottom-right coordinates of bounding boxes.
        Uses formula: area = width × height
    
    Input:
        box (torch.Tensor): Bounding box coordinates in (x1, y1, x2, y2) format
                           with shape (4,) or (4, n)
    
    Output:
        torch.Tensor: Bounding box area
    
    Example:
        >>> box = torch.tensor([10, 10, 50, 30])  # (x1, y1, x2, y2)
        >>> area = box_area(box)
        >>> print(area)  # 800 (width=40 × height=20)
    """
    # box = xyxy(4,n)
    return (box[2] - box[0]) * (box[3] - box[1])



def box_iou(box1, box2, eps=1e-7):
    """
    Calculate Intersection over Union (IoU) values between bounding boxes.
    
    Function:
        Calculate IoU (Intersection over Union) values between two sets of bounding boxes.
        Also known as Jaccard index, used to measure the overlap between two bounding boxes.
    
    Input:
        box1 (torch.Tensor): First set of bounding boxes with shape (N, 4) in (x1, y1, x2, y2) format
        box2 (torch.Tensor): Second set of bounding boxes with shape (M, 4) in (x1, y1, x2, y2) format
        eps (float): Small value to prevent division by zero, default 1e-7
    
    Output:
        torch.Tensor: IoU matrix with shape (N, M)
                     containing IoU values for each pair of bounding boxes from box1 and box2
    
    Example:
        >>> box1 = torch.tensor([[0, 0, 10, 10]])  # one bounding box
        >>> box2 = torch.tensor([[5, 5, 15, 15]])  # another bounding box
        >>> iou = box_iou(box1, box2)
        >>> print(iou)  # [[0.1429]] (intersection area 25 / union area 175)
    """
    # https://github.com/pytorch/vision/blob/master/torchvision/ops/boxes.py

    # inter(N,M) = (rb(N,M,2) - lt(N,M,2)).clamp(0).prod(2)
    (a1, a2), (b1, b2) = box1[:, None].chunk(2, 2), box2.chunk(2, 1)
    inter = (torch.min(a2, b2) - torch.max(a1, b1)).clamp(0).prod(2)

    # IoU = inter / (area1 + area2 - inter)
    return inter / (box_area(box1.T)[:, None] + box_area(box2.T) - inter + eps)
    

def non_max_suppression(
    prediction,
    conf_threshold=0.25,
    iou_threshold=0.7,
    classes=None,
    agnostic=False,
    max_det=300,
    nc=0,
    version='yolo11',
    task=None
):
    """
    Non-Maximum Suppression (NMS) processing on inference results to remove overlapping detections.
    
    Function:
        Apply Non-Maximum Suppression to YOLO model inference results.
        Remove low-confidence and high-overlap detection boxes, retaining the best detections.
        Supports multiple YOLO versions (yolov5, yolov8, yolo11, etc.).
    
    Input:
        prediction (torch.Tensor | numpy.ndarray): Model inference results
        conf_thres (float): Confidence threshold, default 0.25
        iou_thres (float): IoU threshold, default 0.7
        classes (list, optional): Specific class indices to keep
        agnostic (bool): Whether to perform class-agnostic NMS, default False
        max_det (int): Maximum number of detections per image, default 300
        nc (int): Number of classes, default 0 (auto-calculate)
        version (str): YOLO version, default 'yolov11'
    
    Output:
        list: List of detection results for each image
             Each element is a tensor with shape (n, 6) in format [x1, y1, x2, y2, conf, cls]
    
    Example:
        >>> prediction = model(images)
        >>> detections = non_max_suppression(prediction, conf_thres=0.5, iou_thres=0.45)
        >>> for det in detections:
        ...     print(det.shape)  # torch.Size([n, 6])
    """    
    if isinstance(prediction, (list, tuple)):  # YOLOv5 model in validation model, output = (inference_out, loss_out)
        prediction = prediction[0]  # select only inference output
    # Settings
    max_wh = 7680  # (pixels) maximum box width and height
    max_nms = 3000  # maximum number of boxes into torchvision.ops.nms()
    num_bbox_conf = 5 if version == 'yolov5' else 4
    multiplier = 0 if agnostic else max_wh
    v5_task_not_pose = task != 'pose'

    bs, _, num_feat = prediction.shape
    nc = nc or (num_feat - num_bbox_conf)  # e.g. 36
    mi = num_bbox_conf + nc 
    classes_tensor = torch.tensor(classes, device=prediction.device) if classes is not None else None

    time_limit = 0.5 + 0.05 * bs  # seconds to quit after
    output = [torch.zeros((0, 6), device=prediction.device)] * bs

    t = time.time()
    for xi, x in enumerate(prediction):  # image index, image inference
        if version == 'yolov5':
            xc = x[..., 4] > conf_threshold
            if v5_task_not_pose:
                x[..., 5:] *= x[..., 4, None]
        elif version in ['yolov5u', 'yolov8', 'yolo11']:
            x = x.transpose(-1, -2)
            xc = x[..., 4:mi].amax(-1) > conf_threshold
        x = x[xc]  # confidence

        if not x.shape[0]:
            continue
        box  = xywh2xyxy(x[..., :4])
        cls  = x[:, num_bbox_conf:num_bbox_conf+nc]
        mask = x[:, num_bbox_conf+nc:]

        conf, j = cls.max(1, keepdim=True)
        x = torch.cat((box, conf, j.float(), mask), 1)[conf.view(-1) > conf_threshold]
        # Filter by class
        if classes is not None:
            x = x[(x[..., 5:6] == classes_tensor).any(1)]
        # Check shape
        n = x.shape[0]  # number of boxes
        if not n:  # no boxes
            continue

        # Batched NMS
        # boxes, scores = x[..., :4] + c, x[..., 4]  # boxes (offset by class), scores
        # c = x[..., 5:6] * multiplier # c is the class offset if agnostic is True all boxes view as same class, else c is the class offset for each box depending on the class
        # i = nms(boxes + c, scores, iou_threshold=iou_threshold)[:max_det] # NMS 

        i = nms(x[..., :4] + x[..., 5:6] * multiplier, x[..., 4], iou_threshold=iou_threshold)[:max_det] # NMS 
        output[xi] = x[i]
        if (time.time() - t) > time_limit:
            break  # time limit exceeded
    return output


def letterbox(img, new_shape=(640, 640), color=(114, 114, 114)):
    """
    Resize image to specified dimensions while maintaining aspect ratio and padding borders.
    
    Function:
        Resize input image to specified new dimensions while maintaining original aspect ratio.
        Fill empty areas with specified color, ensuring image is a multiple of 32 pixels.
    
    Input:
        img (numpy.ndarray): Input image with shape (height, width, channels)
        new_shape (tuple | int): Target dimensions, default (640, 640)
        color (tuple): Padding color (R, G, B), default (114, 114, 114)
    
    Output:
        numpy.ndarray: Resized image with dimensions new_shape
    
    Example:
        >>> img = cv2.imread('image.jpg')  # shape (480, 640, 3)
        >>> resized = letterbox(img, new_shape=(416, 416))
        >>> print(resized.shape)  # (416, 416, 3)
    """
    # Resize image to a 32-pixel-multiple rectangle https://github.com/ultralytics/yolov3/issues/232
    shape = img.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)
    elif isinstance(new_shape, tuple):
        new_shape = new_shape
    else:
        raise TypeError(f"Unsupported type for new_shape: {type(new_shape)}")
    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    # Compute padding
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    dw /= 2  # divide padding into 2 sides
    dh /= 2
    
    if shape[::-1] != new_unpad:  # resize
        img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    return img

#==========================================
# Classification
#==========================================
DEFAULT_MEAN = (0.0, 0.0, 0.0)
DEFAULT_STD = (1.0, 1.0, 1.0)
DEFAULT_CROP_FRACTION = 1.0
def classify_transforms(
    size=224,
    mean=DEFAULT_MEAN,
    std=DEFAULT_STD,
    interpolation="BILINEAR",
    crop_fraction: float = DEFAULT_CROP_FRACTION,
):
    """
    Create a composition of image transforms for classification tasks.
    
    Function:
        Generate image preprocessing transform sequence suitable for classification models.
        Includes resizing, center cropping, tensor conversion, and normalization.
    
    Input:
        size (int | tuple): Target image size
                          If integer, represents shortest edge length
                          If tuple, represents (height, width)
        mean (tuple): RGB channel mean values for normalization
        std (tuple): RGB channel standard deviation values for normalization
        interpolation (str): Interpolation method, options: 'NEAREST', 'BILINEAR', or 'BICUBIC'
        crop_fraction (float): Crop fraction, default DEFAULT_CROP_FRACTION
    
    Output:
        torchvision.transforms.Compose: Composition of image transforms
    
    Example:
        >>> transforms = classify_transforms(size=224)
        >>> img = Image.open("path/to/image.jpg")
        >>> transformed_img = transforms(img)
        >>> print(transformed_img.shape)  # torch.Size([3, 224, 224])
    """
    import torchvision.transforms as T  # scope for faster 'import ultralytics'

    if isinstance(size, (tuple, list)):
        assert len(size) == 2, f"'size' tuples must be length 2, not length {len(size)}"
        scale_size = tuple(math.floor(x / crop_fraction) for x in size)
    else:
        scale_size = math.floor(size / crop_fraction)
        scale_size = (scale_size, scale_size)

    # Aspect ratio is preserved, crops center within image, no borders are added, image is lost
    if scale_size[0] == scale_size[1]:
        # Simple case, use torchvision built-in Resize with the shortest edge mode (scalar size arg)
        tfl = [T.Resize(scale_size[0], interpolation=getattr(T.InterpolationMode, interpolation))]
    else:
        # Resize the shortest edge to matching target dim for non-square target
        tfl = [T.Resize(scale_size)]
    tfl.extend(
        [
            T.CenterCrop(size),
            T.ToTensor(),
            T.Normalize(mean=torch.tensor(mean), std=torch.tensor(std)),
        ]
    )
    return T.Compose(tfl)

#==========================================
# Mask
#==========================================
def process_mask(protos, masks_in, bboxes, shape, upsample=False, version='yolo11'):
    """
    Apply masks to bounding boxes using the output of the mask head.
    
    Function:
        Combine prototype masks and mask coefficients to generate final segmentation masks.
        Supports different YOLO model versions and optional upsampling.
    
    Input:
        protos (torch.Tensor): Prototype mask tensor with shape [mask_dim, mask_h, mask_w]
        masks_in (torch.Tensor): Mask coefficients tensor with shape [n, mask_dim]
                                where n is the number of masks after NMS
        bboxes (torch.Tensor): Bounding box tensor with shape [n, 4]
        shape (tuple): Input image dimensions (h, w)
        upsample (bool): Whether to upsample masks to original image size, default False
        version (str): YOLO version, default 'yolo11'
    
    Output:
        torch.Tensor: Binary mask tensor with shape [n, h, w]
                     where n is the number of masks after NMS
    
    Example:
        >>> protos = torch.rand(32, 160, 160)  # prototype masks
        >>> masks_in = torch.rand(5, 32)  # mask coefficients for 5 objects
        >>> bboxes = torch.rand(5, 4)  # 5 bounding boxes
        >>> shape = (640, 640)  # image dimensions
        >>> masks = process_mask(protos, masks_in, bboxes, shape)
        >>> print(masks.shape)  # torch.Size([5, 640, 640])
    """
    c, mh, mw = protos.shape  # CHW
    ih, iw = shape
    if masks_in.shape[0] == 0:
        return torch.zeros((0, mh, mw), device=protos.device)
    masks = (masks_in @ protos.float().view(c, -1)).view(-1, mh, mw)  # CHW
    width_ratio = mw / iw
    height_ratio = mh / ih

    downsampled_bboxes = bboxes.clone()
    downsampled_bboxes[:, 0] *= width_ratio
    downsampled_bboxes[:, 2] *= width_ratio
    downsampled_bboxes[:, 3] *= height_ratio
    downsampled_bboxes[:, 1] *= height_ratio

    masks = crop_mask(masks, downsampled_bboxes)  # CHW
    if upsample:
        masks = F.interpolate(masks[None], shape, mode="bilinear", align_corners=False)[0]  # CHW
    if version in ['yolov8', 'yolo11']:
        threshold = 0.0
    elif version == 'yolov5':
        threshold = 0.5
    else:
        raise ValueError(f"Invalid version: {version}")
    return masks.gt_(threshold)


def crop_mask(masks, boxes):
    """
    Crop predicted masks to bounding box regions.
    
    Function:
        Set mask pixels outside bounding boxes to zero, retaining masks within bounding boxes.
        Uses vectorized operations for improved efficiency.
    
    Input:
        masks (torch.Tensor): Mask tensor with shape [n, h, w]
        boxes (torch.Tensor): Bounding box coordinate tensor with shape [n, 4]
                             in relative coordinate format (x1, y1, x2, y2)
    
    Output:
        torch.Tensor: Cropped masks with shape [n, h, w]
    
    Example:
        >>> masks = torch.rand(2, 100, 100)  # 2 masks
        >>> boxes = torch.tensor([[10, 10, 50, 50], [20, 20, 80, 80]])  # 2 bounding boxes
        >>> cropped = crop_mask(masks, boxes)
        >>> print(cropped.shape)  # torch.Size([2, 100, 100])
    """
    _, h, w = masks.shape
    x1, y1, x2, y2 = torch.chunk(boxes[:, :, None], 4, 1)  # x1 shape(1,1,n)
    r = torch.arange(w, device=masks.device, dtype=x1.dtype)[None, None, :]  # rows shape(1,w,1)
    c = torch.arange(h, device=masks.device, dtype=x1.dtype)[None, :, None]  # cols shape(h,1,1)

    return masks * ((r >= x1) * (r < x2) * (c >= y1) * (c < y2))


def scale_masks(masks, shape, padding: bool = True):
    """
    Rescale segmentation masks to target shape.
    
    Function:
        Scale masks from one size to another size.
        Supports YOLO-style padding handling.
    
    Input:
        masks (torch.Tensor): Mask tensor with shape (N, C, H, W)
        shape (tuple): Target height and width (height, width)
        padding (bool): Whether masks are based on YOLO-style padded images, default True
    
    Output:
        torch.Tensor: Rescaled masks
    
    Example:
        >>> masks = torch.rand(1, 1, 160, 160)  # 1 mask
        >>> shape = (640, 640)  # target size
        >>> scaled = scale_masks(masks, shape)
        >>> print(scaled.shape)  # torch.Size([1, 1, 640, 640])
    """
    mh, mw = masks.shape[2:]
    gain = min(mh / shape[0], mw / shape[1])  # gain  = old / new
    pad = [mw - shape[1] * gain, mh - shape[0] * gain]  # wh padding
    if padding:
        pad[0] /= 2
        pad[1] /= 2
    top, left = (int(round(pad[1] - 0.1)), int(round(pad[0] - 0.1))) if padding else (0, 0)  # y, x
    bottom, right = (
        mh - int(round(pad[1] + 0.1)),
        mw - int(round(pad[0] + 0.1)),
    )
    masks = masks[..., top:bottom, left:right]

    masks = F.interpolate(masks, shape, mode="bilinear", align_corners=False)  # NCHW
    return masks
    

def scale_image(masks, im0_shape, ratio_pad=None):
    """
    Scale masks or images to original image dimensions.
    
    Function:
        Scale resized and padded masks or images back to original image dimensions.
        Remove padding and adjust to correct dimensions.
    
    Input:
        masks (np.ndarray): Resized and padded masks/images
                           with shape [h, w, num] or [h, w, 3]
        im0_shape (tuple): Original image shape (height, width)
        ratio_pad (tuple, optional): Padding ratio and padding values
                                   in format (ratio, pad)
    
    Output:
        np.ndarray: Scaled masks with shape [h, w, num]
    
    Example:
        >>> masks = np.random.rand(640, 640, 3)  # padded image
        >>> im0_shape = (480, 640)  # original image size
        >>> scaled = scale_image(masks, im0_shape)
        >>> print(scaled.shape)  # (480, 640, 3)
    """
    # Rescale coordinates (xyxy) from im1_shape to im0_shape
    im1_shape = masks.shape
    if im1_shape[:2] == im0_shape[:2]:
        return masks
    if ratio_pad is None:  # calculate from im0_shape
        gain = min(im1_shape[0] / im0_shape[0], im1_shape[1] / im0_shape[1])  # gain  = old / new
        pad = (im1_shape[1] - im0_shape[1] * gain) / 2, (im1_shape[0] - im0_shape[0] * gain) / 2  # wh padding
    else:
        # gain = ratio_pad[0][0]
        pad = ratio_pad[1]
    top, left = int(pad[1]), int(pad[0])  # y, x
    bottom, right = int(im1_shape[0] - pad[1]), int(im1_shape[1] - pad[0])

    if len(masks.shape) < 2:
        raise ValueError(f'"len of masks shape" should be 2 or 3, but got {len(masks.shape)}')
    masks = masks[top:bottom, left:right]
    masks = cv2.resize(masks, (im0_shape[1], im0_shape[0]))
    if len(masks.shape) == 2:
        masks = masks[:, :, None]
    return masks

def min_index(arr1, arr2):
    """
    Find a pair of indexes with the shortest distance between two arrays of 2D points.

    Args:
        arr1 (np.ndarray): A NumPy array of shape (N, 2) representing N 2D points.
        arr2 (np.ndarray): A NumPy array of shape (M, 2) representing M 2D points.

    Returns:
        (tuple): A tuple containing the indexes of the points with the shortest distance in arr1 and arr2 respectively.
    """
    dis = ((arr1[:, None, :] - arr2[None, :, :]) ** 2).sum(-1)
    return np.unravel_index(np.argmin(dis, axis=None), dis.shape)

def merge_multi_segment(segments):
    """
    Merge multiple segments into one list by connecting the coordinates with the minimum distance between each segment.
    This function connects these coordinates with a thin line to merge all segments into one.

    Args:
        segments (List[List]): Original segmentations in COCO's JSON file.
                               Each element is a list of coordinates, like [segmentation1, segmentation2,...].

    Returns:
        s (List[np.ndarray]): A list of connected segments represented as NumPy arrays.
    """
    s = []
    segments = [np.array(i).reshape(-1, 2) for i in segments]
    idx_list = [[] for _ in range(len(segments))]

    # Record the indexes with min distance between each segment
    for i in range(1, len(segments)):
        idx1, idx2 = min_index(segments[i - 1], segments[i])
        idx_list[i - 1].append(idx1)
        idx_list[i].append(idx2)

    # Use two round to connect all the segments
    for k in range(2):
        # Forward connection
        if k == 0:
            for i, idx in enumerate(idx_list):
                # Middle segments have two indexes, reverse the index of middle segments
                if len(idx) == 2 and idx[0] > idx[1]:
                    idx = idx[::-1]
                    segments[i] = segments[i][::-1, :]

                segments[i] = np.roll(segments[i], -idx[0], axis=0)
                segments[i] = np.concatenate([segments[i], segments[i][:1]])
                # Deal with the first segment and the last one
                if i in {0, len(idx_list) - 1}:
                    s.append(segments[i])
                else:
                    idx = [0, idx[1] - idx[0]]
                    s.append(segments[i][idx[0] : idx[1] + 1])

        else:
            for i in range(len(idx_list) - 1, -1, -1):
                if i not in {0, len(idx_list) - 1}:
                    idx = idx_list[i]
                    nidx = abs(idx[1] - idx[0])
                    s.append(segments[i][nidx:])
    return s

def masks2segments(masks, strategy: str = "all"):
    """
    Convert masks to segments using contour detection.

    Args:
        masks (torch.Tensor): Binary masks with shape (batch_size, 160, 160).
        strategy (str): Segmentation strategy, either 'all' or 'largest'.

    Returns:
        (list): List of segment masks as float32 arrays.
    """
    segments = []
    for x in masks.int().cpu().numpy().astype("uint8"):
        c = cv2.findContours(x, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[0]
        if c:
            if strategy == "all":  # merge and concatenate all segments
                c = (
                    np.concatenate(merge_multi_segment([x.reshape(-1, 2) for x in c]))
                    if len(c) > 1
                    else c[0].reshape(-1, 2)
                )
            elif strategy == "largest":  # select largest segment
                c = np.array(c[np.array([len(x) for x in c]).argmax()]).reshape(-1, 2)
        else:
            c = np.zeros((0, 2))  # no segments found
        segments.append(c.astype("float32"))
    return segments




from typing import List, Tuple, Optional
from PIL import Image

import numpy as np
import cv2

from .base_predictor import BasePredictor
from ..utils import np_ops


class ClassificationPredictor(BasePredictor):
    def __init__(self, config: dict) -> None:
        super().__init__(config)
        
        self.transforms = np_ops.classify_transforms(size=self.input_shape)

    def preprocess(self, imgs: List[np.ndarray]) -> Tuple[np.ndarray, List[np.ndarray]]:
        """Preprocess images for classification."""
        img_stack = np.stack(
            [self.transforms(Image.fromarray(cv2.cvtColor(im, cv2.COLOR_BGR2RGB))) for im in imgs]
        )
        return img_stack, imgs

    def postprocess(
        self, 
        preds: np.ndarray, 
        orig_imgs: Optional[List[np.ndarray]] = None
    ) -> Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        preds = preds[0] if isinstance(preds, (list, tuple)) else preds
        top5_idx = np.argsort(preds, axis=1)[:, -5:][:, ::-1]
        top5 = np.take_along_axis(preds, top5_idx, axis=1)
        return preds, (top5, top5_idx)


import os
import gc
import cv2
cv2.setNumThreads(0)
import time
import yaml
import uuid
import logging
import platform
import threading
import numpy as np

from .utils import is_jetson

# Obtain the "nxva.streaming" logger, which will naturally propagate to the "nxva" logger
_logger = logging.getLogger(__name__)
_logger.addHandler(logging.NullHandler())  # Avoid the "No handler" warning


class MultiStreaming(object):
    """
    A class to handle multi-camera streaming with automatic reconnection, frame resizing and gstreamer support.

    Attributes:
        reconnect_interval (int): Interval in seconds to attempt reconnection to a camera.
        logger (Logger): Logger instance for logging messages.
        cfg (dict): Configuration settings for the cameras which read from a YAML file or dictionary.
        w (int): Width of the frame.
        h (int): Height of the frame.
        gst_enable (bool): Whether to use GStreamer for streaming.
        codec (str): Codec to use for streaming, default is 'h264'.
        cameras (dict): Dictionary containing camera configurations, where each key is a unique item_id 
                        and the value is a dictionary with the following keys:
            - item_id (str): Unique identifier for the camera.  Generated by uuid if not provided.
            - rtsp (list): List of dictionaries containing RTSP stream information:
                - ip (str): IP address of the camera.
                - user (str, optional): Username for the camera.
                - pw (str, optional): Password for the camera.
            - codec (str): Codec to use to generate gstreamer pipeline, default is 'h264'.
            - status (bool): Whether the camera is currently active.
            - rtsp_strings (list): List of RTSP string candidates
            - rtsp_reconnect (str): RTSP string to use for reconnection.
            - cap (cv2.VideoCapture): OpenCV VideoCapture object for the camera.
            - cap_info (dict): Dictionary containing camera properties like width, height, FPS, and codec.
        num (int): Number of cameras configured.
        running (bool): Flag to indicate if the streaming is running.
        mutex (threading.Lock): Mutex for thread-safe access to camera frames.
        threads (list): List of threads for each camera stream.
        frames (list): List of frames from each camera.
        id_to_idx (dict): Mapping from camera item_id to index in the frames list.
    """
    def __init__(
        self, 
        config=None, 
        reconnect_interval=30,
        logger=None,
        verbose=True
    ):
        """
        Initialize the MultiStreaming object with the given configuration settings.
        
        Args:
            config (str or dict): Configuration settings for the cameras, 
                which can be provided as a path to a YAML file or directly as a dictionary. 
                The configuration should include the following sections:
                
                'parameter': Dictionary containing:
                    - 'WIDTH' (int): Width of the frame.
                    - 'HEIGHT' (int): Height of the frame.
                    - 'GSTREAMER' (bool): Whether to use GStreamer for streaming.
                    - 'CODEC' (str, optional): Global codec to use for gstreamer
                
                'cameras': List of dictionaries, each containing:
                    - 'item_id' (str, optional): Unique identifier for the camera. If not provided, a UUID will be generated.
                    - 'ip' (str): IP address of the camera.
                    - 'user' (str, optional): Username for the camera.
                    - 'pw' (str, optional): Password for the camera.
                
                Example configuration:
                ```
                parameter:
                    WIDTH: 1920
                    HEIGHT: 1080
                    GSTREAMER: True
                    CODEC: h264
                cameras:
                    - ip: '192.168.1.100'
                      user: 'admin'
                      pw: 'password'
                ```
                or 
                ```
                cameras:
                    - item_id: 'CAMERA-1'
                      rtsp:
                        - ip: '192.168.1.100'
                          user: 'admin'
                          pw: 'password'
                      codec: 'h264'
                ```
            reconnect_interval (int): Interval in seconds to attempt reconnection to a camera.
            verbose (bool): Enable verbose output to help with debugging.
        """
        self.reconnect_interval = reconnect_interval
        if not verbose:
            _logger.setLevel(logging.ERROR)
        self.logger = logger or _logger
        
        if config:
            self._load_and_validate_config(config)
            self._generate_rtsp_strings()
        else:
            raise ValueError('Configuration must be provided, either as a dictionary or a YAML file path.')
        
        self.running = True
        self.mutex = threading.Lock()
        self.threads = []  # store streaming threads for each camera
        self.frames = [np.zeros((self.h, self.w, 3), dtype=np.uint8) for i in range(self.num)]
        self.id_to_idx = {cam['item_id']: i for i, cam in enumerate(self.cameras.values())}

    def _load_and_validate_config(self, config):
        if isinstance(config, str):
            assert config.endswith(('.yaml', '.yml'))
            with open(config, 'r') as f:
                self.cfg = yaml.safe_load(f)
        elif isinstance(config, dict):
            self.cfg = config
        else:
            raise TypeError('Config must be provided as a yaml file or dictionary.')
        
        # Ensure 'cameras' section exists and is a list
        if 'cameras' not in self.cfg or not isinstance(self.cfg['cameras'], list):
            raise ValueError("The 'cameras' section must be a list and cannot be empty.")
        
        param = self.cfg.get('parameter', {}) or {}
        self.w = int(param.get('WIDTH', 1920))
        self.h = int(param.get('HEIGHT', 1080))
        self.candidate = bool(param.get('CANDIDATE', False))
        self.gst_enable = bool(param.get('GSTREAMER', False))
        self.backend = param.get('BACKEND', None)  # 'opencv', 'gst' or 'ffmpeg'
        # Backward compatibility: if backend is not set, use 'gst' if GStreamer is enabled
        if self.backend is None:
            self.backend = 'gst' if self.gst_enable else 'opencv'
        self.codec = param.get('CODEC', 'h264')  # global setting for codec
        # FFmpeg exclusive parameters (optional)
        ffmpeg_param = param.get('FFMPEG', {}) or {}
        self.ffmpeg_fps = int(ffmpeg_param.get('FPS', 5))
        self.ffmpeg_hwaccel = bool(ffmpeg_param.get('HWACCEL', True))
        self.ffmpeg_fallback = bool(ffmpeg_param.get('FALLBACK', False))
        self.logger.info(f"Streaming global settings: width={self.w}, height={self.h}, backend={self.backend}, codec={self.codec}, "
                         f"ffmpeg_fps={self.ffmpeg_fps}, ffmpeg_hwaccel={self.ffmpeg_hwaccel}, ffmpeg_fallback={self.ffmpeg_fallback}")

        self.cameras = {}
        for cam in self.cfg['cameras']:
            item_id = cam.get('item_id', str(uuid.uuid4()))
            cam['item_id'] = item_id  # confirm item_id is set
            if item_id in self.cameras:
                raise ValueError(f"Duplicate item_id found: {item_id}. Each camera must have a unique item_id.")
            
            # Ensure global codec in camera configuration
            # If codec is not specified, use the global codec
            if 'codec' not in cam:
                cam['codec'] = self.codec
            cam['codec'] = cam['codec'].lower()
            if cam['codec'] not in ['h264', 'h265']:
                raise ValueError(f"Unsupported codec '{cam['codec']}' for camera {item_id}. Supported codecs are: h264, h265.")
            
            if 'rtsp' in cam:
                if not isinstance(cam['rtsp'], list):
                    raise ValueError(f"Camera {item_id} 'rtsp' must be a list of dictionary with 'ip', 'user', and 'pw'.")
                else:
                    # Validate each RTSP entry
                    for rtsp in cam['rtsp']:
                        if not isinstance(rtsp, dict):
                            raise ValueError(f"Each RTSP entry for camera {item_id} must be a dictionary with 'ip', 'user', and 'pw'.")
                        if 'ip' not in rtsp:
                            raise ValueError(f"RTSP entry for camera {item_id} must have 'ip' defined.")
                        if 'user' not in rtsp:
                            rtsp['user'] = None
                        if 'pw' not in rtsp:
                            rtsp['pw'] = None
                        if 'codec' not in rtsp:
                            rtsp['codec'] = cam['codec']
                        rtsp['codec'] = rtsp['codec'].lower()
                        if rtsp['codec'] not in ['h264', 'h265']:
                            raise ValueError(f"Unsupported codec '{rtsp['codec']}' for camera {item_id}. Supported codecs are: h264, h265.")
            else:
                ip = cam.get('ip', None)
                if not ip:
                    raise ValueError(f"Camera {item_id} must have 'ip' or 'rtsp' defined.")
                cam['rtsp'] = [{'ip': ip, 'user': cam.get('user', None), 'pw': cam.get('pw', None), 'codec': cam['codec']}]
            
            cam['status'] = False
            cam['rtsp_strings'] = []
            cam['rtsp_reconnect'] = None
            cam['cap'] = cv2.VideoCapture()
            cam['cap_info'] = {}
            self.cameras[item_id] = cam

        self.num = len(self.cameras)

    def _generate_rtsp_strings(self):
        cpu_arch = platform.machine()
        # Collect RTSP strings
        for item_id, cam in self.cameras.items():
            for rtsp in cam['rtsp']:
                ip = rtsp['ip']
                user = rtsp['user']
                pw = rtsp['pw']
                codec = rtsp['codec']

                if self.backend == 'gst':
                    gst_pipeline = self._gst_pipeline_template(cpu_arch, need_user_pw=bool(user and pw))
                    if gst_pipeline:
                        rtsp_string = gst_pipeline.format(ip=ip, user=user, pw=pw, codec=codec)
                        # Gstreamer pipeline put in front of the default rtsp
                        if len(cam['rtsp_strings']) > 1:
                            index = len(cam['rtsp_strings']) // 2
                            cam['rtsp_strings'].insert(index, rtsp_string)
                        else:
                            cam['rtsp_strings'].append(rtsp_string)
                
                # if Gstreamer is not used, use the default rtsp
                # if Gstreamer is used, this will be backup
                if user and pw:
                    cam['rtsp_strings'].append(f'rtsp://{user}:{pw}@{ip}')
                else:
                    cam['rtsp_strings'].append(f'rtsp://{ip}')

        log_string = (
            f"RTSP strings for {len(self.cameras)} cameras:\n"
            + "\n".join(f"{item_id}: {cam['rtsp_strings']}" for cam in self.cameras.values())
        )
        self.logger.debug(log_string)

    def _gst_pipeline_template(self, cpu_arch, need_user_pw=False):
        rtspsrc_base  = 'rtspsrc location=rtsp://{ip}'
        if need_user_pw:
            rtspsrc_base += ' user-id={user} user-pw={pw}'

        # https://forums.developer.nvidia.com/t/deepstream-losing-relevance-because-of-outdated-dependecies/203426/7?utm_source=chatgpt.com 
        rtspsrc_params = [
            'short-header=true',
            'is-live=true',
            'protocols=tcp',
            'user-agent="LibVLC/3.0.18"',
            'do-rtsp-keep-alive=true',
            'timeout=5000000',  # 5s
            'tcp-timeout=5000000'
            # 'latency=200',
            # 'drop-on-latency=true',
            # 'timeout=10',
            # 'tcp-timeout=5',
            # 'retry=3',
            # 'do-retransmission=false'
        ]
        rtspsrc = rtspsrc_base + ' ' + ' '.join(rtspsrc_params)
        
        if cpu_arch == 'x86_64':
            decode_chain = [
                f'queue max-size-buffers=0 max-size-bytes={1024*1024} leaky=downstream',  # 1MB buffer to keep GOP not be truncated
                'rtp{codec}depay', 
                '{codec}parse', 
                'avdec_{codec}', 
                'videoconvert n-threads=1', 
                'appsink drop=true sync=false'
            ]
            # decode_chain = [
            #     f'queue max-size-buffers=0 max-size-bytes={1024*1024} leaky=downstream',
            #     'rtp{codec}depay', 
            #     '{codec}parse', 
            #     'nv{codec}dec',
            #     'nvvideoconvert'  # it is important to use nvvideoconvert after nv{codec}dec for performance
            #     'queue max-size-buffers=1 leaky=downstream',
            #     'videoconvert n-threads=1',  # it is important to add convert after nvvideoconvert for performance
            #     'video/x-raw, format=RGBA',
            #     'appsink caps="video/x-raw,format=BGR drop=true sync=false'
            # ]
        elif cpu_arch == 'aarch64':
            if is_jetson():
                decode_chain = [
                    f'queue max-size-buffers=0 max-size-bytes={1024*1024} leaky=downstream',
                    'rtp{codec}depay', 
                    '{codec}parse config-interval=1', 
                    'nvv4l2decoder',  #  drop-frame has some problem, disable-dpb=true need the stream with no B-frames
                    'nvvidconv', 
                    'queue max-size-buffers=1 leaky=downstream',
                    'video/x-raw,format=RGBA',          # explicitly set RGBA format for system memory
                    'videoconvert n-threads=1',         
                    'video/x-raw,format=BGR',           # cap filter, ensure the output is in BGR format
                    'appsink drop=true sync=false'
                ]
                # decode_chain = [
                #     'rtp{codec}depay', '{codec}parse', 'nvv4l2decoder', 
                #     'nvvidconv', 'appsink drop=true sync=false'
                # ]
            else:
                decode_chain = [
                'rtp{codec}depay', 
                '{codec}parse', 
                'avdec_{codec}', 
                'videoconvert n-threads=1', 
                'appsink drop=true sync=false'
            ]
        else:
            decode_chain = []
            self.logger.warning('Not allow to use Gstreamer under the current platform! (Only support x86_64 and Jetson)')

        if not decode_chain:
            return ''
        return ' ! '.join([rtspsrc] + decode_chain)
    
    def __len__(self):
        return self.num
    
    def run(self):
        """Start the camera streaming threads."""
        # for i in range(self.num):
        for cam in self.cameras.values():
            t = threading.Thread(
                target=self._run_one,
                args=(cam, ),
                daemon=True
                )
            t.start()
            self.threads.append(t)

    def get_frames(self):
        """
        Retrieve the latest frames from all cameras.
        """
        with self.mutex:
            frames = self.frames.copy()
            return frames
        
    def get_status(self):
        """
        Retrieve the status of all cameras.
        Returns a list of booleans indicating whether each camera is active.
        """
        with self.mutex:
            status_list = [cam['status'] for cam in self.cameras.values()]
            return status_list
        
    def stop(self):
        """Stop all camera threads."""
        self.running = False

    def release(self):
        """Stop all camera threads and release resources."""
        self.running = False
        time.sleep(0.1)
        for cam in self.cameras.values():
            if 'cap' in cam and cam['cap'].isOpened():
                cam['cap'].release()

        # Wait for threads to finish
        for thread in self.threads:
            thread.join()
        self.threads.clear()
        self.logger.info("All camera threads stopped and resources released.")
            
    def _run_one(self, cam):
        """Handle the streaming of one individual camera."""
        self._init_camera(cam)
        cap = cam['cap']
        resize = (cam['cap_info']['w'], cam['cap_info']['h']) != (self.w, self.h)
        reconnect_time = time.time() - self.reconnect_interval
        
        try:
            fail = 0
            while self.running:
                # successfully read frame
                if cap.isOpened():
                    ret, frame = cap.read()
                    if ret:
                        if not cam['status']:
                            self.logger.info(f"Camera {cam['item_id']} reconnects successfully with RTSP \"{cam['rtsp_reconnect']}\"")
                        if resize:
                            frame = cv2.resize(frame, (self.w, self.h))
                        with self.mutex:
                            self.frames[self.id_to_idx[cam['item_id']]] = frame
                            cam['status'] = True
                        time.sleep(0.02)
                        continue
                    else:
                        fail += 1
                else:
                    fail += 1
                
                # camera is not open or failed to read frame
                with self.mutex:
                    self.frames[self.id_to_idx[cam['item_id']]] = np.zeros((self.h, self.w, 3), dtype=np.uint8)
                    cam['status'] = False
                    
                if time.time() - reconnect_time > self.reconnect_interval and fail > 10:
                    reconnect_time = time.time()
                    self.logger.warning(f"Camera {cam['item_id']} is not open, try reconnecting with RTSP \"{cam['rtsp_reconnect']}\"")
                    if cap.isOpened():
                        cap.release()
                    del cap
                    gc.collect()
                    time.sleep(0.2)  
                    if cam['rtsp_reconnect']:
                        cap = self._open_camera(cam['rtsp_reconnect'])
                        if not cap.isOpened():
                            if self.candidate:
                                cam['rtsp_reconnect'] = None
                    else:
                        for rtsp in cam['rtsp_strings']:
                            cap = self._open_camera(rtsp)
                            if cap.isOpened():
                                cam['rtsp_reconnect'] = rtsp
                                break
                    cam['cap'] = cap
                    cam['cap_info'] = self._get_info(cam['cap'])
                    resize = (cam['cap_info']['w'], cam['cap_info']['h']) != (self.w, self.h)
                    fail = 0
                
                time.sleep(0.02)
        finally:
            if cap.isOpened():
                cap.release()
    
    def _init_camera(self, cam):
        """Initialize all cameras based on the configuration settings."""
        for rtsp in cam['rtsp_strings']:
            cam['cap'] = self._open_camera(rtsp)
            cam['cap_info'] = self._get_info(cam['cap'])
            
            if cam['cap'].isOpened():
                cam['status'] = True
                cam['rtsp_reconnect'] = rtsp
                self.logger.info(f"Camera {cam['item_id']}: opened successfully with")
                self.logger.info(f"    RTSP: {rtsp}")
                self.logger.info(f"    Info: {cam['cap_info']}")
                break
            else:
                cam['status'] = False
                cam['rtsp_reconnect'] = None if self.candidate else rtsp
                self.logger.warning(f"Camera {cam['item_id']}: failed to open initially with RTSP \"{rtsp}\"")
            
            if not self.candidate:
                # If not candidate mode, break after the first RTSP string
                break

    def _open_camera(self, rtsp):
        """Build capture object based on the backend and RTSP string."""
        if self.backend == 'ffmpeg':
            # Only accept 'rtsp://' strings.
            # If a GStreamer pipeline is passed, it will be ignored and replaced by the upper layer.
            if not rtsp.startswith('rtsp://'):
                cap = type('DummyCap', (), {'isOpened': lambda s: False})()
                return cap
            from .ffmpeg_backend import FFmpegCapture
            cap = FFmpegCapture(self.w, self.h, fps=self.ffmpeg_fps, hwaccel=self.ffmpeg_hwaccel, fallback=self.ffmpeg_fallback)
            cap.open(rtsp)
            return cap
        else:
            cap = self._open_cv2_cap(rtsp, gst=(self.backend == 'gst'))
            return cap
    
    @staticmethod
    def _open_cv2_cap(rtsp, gst=False):
        """Open a camera using OpenCV VideoCapture with optional GStreamer support."""
        cap = cv2.VideoCapture()
        if gst and rtsp.startswith('rtspsrc'):
            t = threading.Thread(
                target=lambda c, u: c.open(u, cv2.CAP_GSTREAMER), 
                args=(cap, rtsp), 
                daemon=True
            )
        else:
            t = threading.Thread(
                target=lambda c, u: c.open(u), 
                args=(cap, rtsp), 
                daemon=True
            )
        t.start()
        t.join(15)
        return cap
    
    def _get_info(self, cap):
        """Return camera properties such as width, height, FPS, and codec."""
        try:
            if cap.isOpened():
                w = cap.get(cv2.CAP_PROP_FRAME_WIDTH)
                h = cap.get(cv2.CAP_PROP_FRAME_HEIGHT)
                fps = cap.get(cv2.CAP_PROP_FPS)
                fourcc = int(cap.get(cv2.CAP_PROP_FOURCC))
                codec_format = ''.join([chr((fourcc >> 8 * i) & 0xFF) for i in range(4)])
                return {
                    'w': int(w),
                    'h': int(h),
                    'fps': int(fps),
                    'codec': codec_format
                }
        except Exception as e:
            self.logger.exception(f"Error getting camera properties: {e}")

        return {
            'w': 0,
            'h': 0,
            'fps': 0,
            'codec': ''
        }


from typing import List, Dict, Any, Tuple, Union

import cv2
import numpy as np

from .detect import DetectionPredictor
from ..utils import np_ops


class SegmentationPredictor(DetectionPredictor):
    """
    A class extending the DetectionPredictor class for prediction based on a segmentation model.
    """
    def __init__(self, config: dict) -> None:
        super().__init__(config)
        self.weight_type = self.config['weights'].split('.')[-1]

    def postprocess(
        self, 
        preds: Union[List[np.ndarray], Tuple[np.ndarray, ...]], 
        orig_imgs: List[np.ndarray]
    ) -> List[Dict[str, Any]]:
        """
        Applies non-max suppression and processes detections for each image in an input batch.
        """
        if self.weight_type == 'engine':
            preds[0], preds[1] = preds[1], preds[0]
        proto = preds[1][None, ...] if preds[1].ndim == 3 else preds[1]

        preds = np_ops.non_max_suppression(
            preds[0],
            conf_threshold=self.conf,
            iou_threshold=self.iou,
            nc=self.nc,
            classes=self.classes,
            agnostic=self.agnostic,
            version=self.version
        )
        results = []
        for i, (pred, orig_img) in enumerate(zip(preds, orig_imgs)):
            if len(pred) == 0:
                results.append({'boxes': np.empty((0, 6)), 'mask': None})
                continue

            masks = np_ops.process_mask(proto[i], pred[:, 6:], pred[:, :4], self.input_shape, upsample=True)  # HWC
            pred[:, :4] = np_ops.scale_boxes(self.input_shape, pred[:, :4], orig_img.shape)
            results.append({'boxes': pred[:, :6], 'mask': masks})
        return results


import time, math
from collections import deque
from dataclasses import dataclass
from typing import Tuple, Dict, Iterable, Deque, Optional

import kp
import numpy as np

from .device import KneronDevice
from .decode_head import decode_yolo_head


class InferenceError(RuntimeError):
    """Raised when any step of the inference pipeline fails."""


@dataclass
class ModelMeta:
    """
    Metadata for a model loaded from NEF.
    Attributes:
        model_id: Unique identifier for the model.
        c: Number of channels in the input tensor.
        h: Height of the input tensor.
        w: Width of the input tensor.
        radix: Radix used for quantization.
        scale: Scale factor used for quantization.
    """
    model_id: int
    c: int
    h: int
    w: int
    radix: int
    scale: float


class InferenceSession:
    """
    Hight-level wrapper that completes a **single‑model** inference workflow on a Kneron USB dongle.
    The class encapsulates the boilerplate steps shown in *plus_python* docs:

    1. Connect to the device and (if necessary) load firmware
    2. Load NEF model
    3. Rearrange layout (4W4C8B) and quantize with radix/scale
    4. Call *generic_data_inference_send / receive*

    Attributes:
        dev: KneronDevice instance, must be connected before use.
        max_inflight: Maximum number of inflight inference requests.
        _model_desc: Loaded model descriptor, if any.
        _models: Dictionary mapping model IDs to ModelMeta instances.
        _inflight: Current number of inflight requests.
        _send_ts: Timestamps of sent requests, used for latency measurement.
        _seq_queue: Sequence numbers of sent requests, used for matching responses.
        _seq_counter: Monotonic sequence number for requests, wraps around at 2^31.
        _model_queue: Queue to match which model the output belongs to.
        _shape_queue: Queue to match original image shape for post-processing.
    """

    #: KL520 / KL720 both support 4W4C8B
    _INPUT_LAYOUT = kp.ModelTensorDataLayout.KP_MODEL_TENSOR_DATA_LAYOUT_4W4C8B
    _WIDTH_ALIGN_BASE = 16
    _CHANNEL_ALIGN_BASE = 4

    def __init__(
        self,
        device: KneronDevice,
        nef_path: str,
        version = 'yolo11',
        task = 'detect',
        max_inflight: int = 2,
    ) -> None:
        self.dev = device
        # self.decode_head = decode_yolo_head
        self.version = version
        self.task = task

        # check the device is connected
        if self.dev.device_group is None:
            self.dev.connect()

        # load model
        if nef_path is not None:
            # self._model_desc = kp.core.load_model_from_file(self.dev.device_group, str(nef_path))
            self.dev.load_model_from_file(str(nef_path))
            self._model_desc = self.dev.model_nef_descriptor
        # else:
        #     # assume already loaded by caller
        #     infos = kp.core.get_system_info(self.dev.device_group)
        #     if not infos or infos[0].number_of_models == 0:
        #         raise InferenceError("No model loaded and nef_path not given")
        #     self._model_desc = None  # we don't actually need it later

        # ------------ collect meta of every model ------------
        self._models: Dict[int, ModelMeta] = {}
        for m in self._model_desc.models:
            node = m.input_nodes[0]
            if node.data_layout != self._INPUT_LAYOUT:
                raise InferenceError(f"Model {m.id} layout {node.data_layout} not 4W4C8B")
            qp = node.quantization_parameters.v1.quantized_fixed_point_descriptor_list[0]
            c, h, w = node.tensor_shape_info.v1.shape_npu[1:]
            self._models[m.id] = ModelMeta(
                model_id=m.id,
                c=c,
                h=h,
                w=w,
                radix=qp.radix,
                scale=qp.scale.value,
            )

        if not self._models:
            raise InferenceError("No model found in NEF")
        
        self.default_model_id = next(iter(self._models))

        # pipeline control
        self.max_inflight = max(max_inflight, 1)
        self._inflight: int = 0     # the number of images sent but not yet received
        self._seq_counter: int = 0  # monotonic inference_number
        self._seq_queue: Deque[int] = deque()
        self._send_ts: Deque[float] = deque()
        self._model_queue: Deque[int] = deque()  # to match which model the output belongs to
        self._shape_queue: Deque[Tuple[int, int]] = deque()  # to match original image shape


    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------
    def _prepare_npu_buffer(self, float_chw: np.ndarray, meta:ModelMeta) -> bytes:
        """
        quantize → 4W4C8B re‑layout → bytes to NPU
        float_chw: (C, H, W) float32 0‑1
        """
        # 1. quantize (radix, scale)
        data = float_chw * (2 ** meta.radix * meta.scale)
        data = np.clip(np.round(data), -128, 127).astype(np.int8)

        # 2. CHW -> HWC
        data = data.transpose(1, 2, 0)  # (H,W,C)

        # 3. 4W4C8B relayout (KL520/KL720 hardware requirement)
        width_align = self._WIDTH_ALIGN_BASE * math.ceil(meta.w / self._WIDTH_ALIGN_BASE)
        ch_block_num = math.ceil(meta.c / self._CHANNEL_ALIGN_BASE)

        re_layout = np.zeros(
            (
                meta.h,
                ch_block_num,
                width_align,
                self._CHANNEL_ALIGN_BASE,
            ),
            dtype=np.int8,
        )

        ch_offset = 0
        for b in range(ch_block_num):
            end = min(ch_offset + self._CHANNEL_ALIGN_BASE, meta.c)
            re_layout[:meta.h, b, :meta.w, :(end - ch_offset)] = data[:, :, ch_offset:end]
            ch_offset += self._CHANNEL_ALIGN_BASE

        return re_layout.tobytes()
    
    def _send_to_npu(self, npu_buffer: bytes, model_id: int) -> None:
        """
        Send a single image to the NPU for inference.
        npu_buffer: image data processed by _prepare_npu_buffer()
        """
        seq = self._seq_counter
        self._seq_counter = (self._seq_counter + 1) & 0x7FFF_FFFF  # wrap around
        desc = kp.GenericDataInferenceDescriptor(
            model_id=model_id,
            inference_number=seq,
            input_node_data_list=[kp.GenericInputNodeData(buffer=npu_buffer)],
        )

        try:
            kp.inference.generic_data_inference_send(
                device_group=self.dev.device_group, 
                generic_inference_input_descriptor=desc
            )
        except kp.ApiKPException as e:
            print(f"Error during inference: {e}")

        self._inflight += 1
        self._send_ts.append(time.time())
        self._seq_queue.append(seq)
        self._model_queue.append(model_id)

    def _recv_from_npu(self):
        try:
            raw = kp.inference.generic_data_inference_receive(
                device_group=self.dev.device_group
            )
        except kp.ApiKPException as e:
            print(f"Error during inference: {e}")

        latency = time.time() - self._send_ts.popleft()
        expected_seq = self._seq_queue.popleft()
        model_id = self._model_queue.popleft()
        orig_shape = self._shape_queue.popleft()

        if raw.header.inference_number != expected_seq:
            raise InferenceError(
                f"inference_number mismatch (expect {expected_seq} got {raw.header.inference_number})"
            )
        
        outputs = [
            kp.inference.generic_inference_retrieve_float_node(
                node_idx=idx,
                generic_raw_result=raw,
                channels_ordering=kp.ChannelOrdering.KP_CHANNEL_ORDERING_CHW,
            ).ndarray
            for idx in range(raw.header.num_output_node)
        ]
        self._inflight -= 1
        return model_id, outputs, latency, orig_shape

    # ---------------------------------------------------------------------
    # Public API
    # ---------------------------------------------------------------------
    def send(self, image: np.ndarray, model_id: Optional[int] = None):
        """
        Asynchronously send a single image for inference.
        Receive one when inflight queue is full.
        """
        model_id = model_id or self.default_model_id
        if model_id not in self._models:
            raise InferenceError(f"model_id {model_id} not in NEF")
        meta = self._models[model_id]
        npu_buf = self._prepare_npu_buffer(image, meta)
        if self._inflight >= self.max_inflight:
            self.recv()
        self._shape_queue.append(image.shape[:2])  # store original shape for postprocessing
        self._send_to_npu(npu_buf, model_id)

    def recv(self, block: bool = True):
        """
        Asynchronously receive inference result.

        Args:
            block: 
                If True, will block until a result is available.
                If False and no result is available, will return None.
        """
        if self._inflight == 0:
            if block:
                raise InferenceError("recv() called with empty inflight queue")
            return None
        model_id, outputs, _, orig_shape = self._recv_from_npu()
        return outputs
    
    def infer(self, image: np.ndarray, model_id: Optional[int] = None):
        """Synchronized inference for a single image."""
        meta = self._models[model_id]
        self.send(image, model_id)
        return self.recv()
        
    def stream(self, images: Iterable[np.ndarray], model_id: Optional[int] = None):
        """Generator that yields inference results for a stream of images."""
        for img in images:
            self.send(img, model_id)
            # receive one result if inflight queue is full
            if self._inflight >= self.max_inflight:
                yield self.recv()
        # receive remaining results
        while self._inflight:
            yield self.recv()

    # ------------------------------------------------------------------
    # Context manager sugar: ensure proper cleanup
    # ------------------------------------------------------------------
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        # Cleanup to avoid resource leaks
        while self._inflight:
            try:
                self.recv()
            except Exception:
                break


import os
import yaml
from typing import List, Dict, Optional
from hailo_sdk_client import ClientRunner
from hailo_sdk_client.exposed_definitions import JoinAction
import numpy as np
from .utils import make_quant_dataset


def load_end_node_config(yaml_path: str = "end_node.yaml") -> Dict:
    """
    Load end_node configuration from YAML file.
    Automatically searches for end_node.yaml in common locations if the default path fails.

    Args:
        yaml_path (str): Path to the YAML configuration file.

    Returns:
        Dict: Configuration dictionary loaded from YAML.
    """
    # List of possible paths to search for end_node.yaml
    possible_paths = [
        yaml_path,  # Use the provided path first
        "end_node.yaml",  # Current directory
        "./end_node.yaml",  # Current directory with explicit path
        "./nxva/nxva/nxhailo/end_node.yaml",  # Relative path from project root
        os.path.join(os.path.dirname(__file__), "end_node.yaml"),  # Same directory as convert.py
        os.path.join(os.path.dirname(os.path.dirname(__file__)), "end_node.yaml"),  # Parent directory
    ]
    
    # Try each possible path
    for path in possible_paths:
        try:
            if os.path.exists(path):
                with open(path, 'r', encoding='utf-8') as file:
                    config = yaml.safe_load(file)
                    print(f"Successfully loaded end_node.yaml from: {path}")
                    return config
        except FileNotFoundError:
            continue
        except yaml.YAMLError as e:
            print(f"Error parsing YAML file {path}: {e}")
            continue
    
    # If no path worked, print warning and return None
    print(f"Warning: Could not find end_node.yaml in any of the searched locations:")
    for path in possible_paths:
        print(f"  - {path}")
    return None

def get_end_node_names_from_dict(convert_table: Dict, model_type: str, task: str, profile_key: str) -> List[str]:
    """
    Get end_node_names for a single model profile from a dictionary.
    Supports both traditional format (YOLOV5/YOLO11) and EMBED format.

    Args:
        convert_table (Dict): Dictionary containing model conversion configurations.
        model_type (str): The model type key in dict (e.g., 'YOLOV5', 'YOLO11', 'EMBED').
        task (str): The task key in dict (e.g., 'detect' for traditional, 'osnet'/'ir'/'par' for EMBED).
        profile_key (str): Profile key (e.g., 'm', 'n') for traditional format, ignored for EMBED format.

    Returns:
        List[str]: end_node_names for the profile.
    """
    # Handle EMBED format (simplified format for models without NMS architecture)
    if model_type == 'EMBED':
        if task not in convert_table[model_type]:
            raise ValueError(f"Task '{task}' not found in EMBED section")
        return convert_table[model_type][task]
    
    # Handle traditional format (YOLOV5/YOLO11)
    if model_type not in convert_table:
        raise ValueError(f"Model type '{model_type}' not found in convert_table")
    
    if task not in convert_table[model_type]:
        raise ValueError(f"Task '{task}' not found for model type '{model_type}'")
    
    task_list = convert_table[model_type][task]
    task_dict = {}
    for entry in task_list:
        task_dict.update(entry)
    
    if profile_key not in task_dict:
        raise ValueError(f"Profile key '{profile_key}' not found in convert_table for {model_type}/{task}")
    
    return task_dict[profile_key]

class MultiModelConvertPipeline:
    """
    ModelConvertPipeline handles multi-model Hailo ONNX conversion, join, normalization, calibration, and compilation.

    Supports automatic join of multiple models, optional insertion of normalization script,
    per-model end_node_names, calibration data handling, and HEF output.
    Configuration is loaded from YAML file (end_node.yaml) or can be provided via custom_end_node_list.

    Attributes:
        model_names (List[str]): List of model names, must match order of end_node_names_list.
        root_dir (str): Root directory for models.
        end_node_names_list (List[List[str]]): List of end_node_names for each model.
        enable_normalization_script (bool): Whether to insert normalization script.
        onnx_paths (List[str]): Internal list of ONNX file paths.
        calib_paths (List[str]): Internal list of calibration data file paths.
        runners (List[ClientRunner]): List of runners for each model.
        r_main (ClientRunner): Main runner after join.
        output_folder (str): Output folder for HEF file.
        convert_table (Dict): Dictionary containing model conversion configurations loaded from YAML.
        custom_calib_paths (Optional[List[str]]): Custom calibration dataset paths for each model.
        custom_end_node_list (Optional[List[List[str]]]): Custom end_node_names for each model.
        end_node_yaml_path (str): Path to the YAML configuration file.
    """
    
    
    def __init__(self, config_dict: Dict):
        """
        Initialize ModelPipeline.

        Args:
            config_dict (Dict): Complete configuration dictionary containing:
                - model_configs: Dict mapping model_name to [model_type, task, profile_key]
                  For traditional models: [model_type, task, profile_key] (e.g., ["YOLOV5", "detect", "m"])
                  For EMBED models: ["EMBED", model_name, ""] (e.g., ["EMBED", "osnet", ""])
                - root_dir: Root directory for models
                - enable_normalization_script: Whether to insert normalization script (optional, default True)
                - custom_calib_paths: Optional list of custom calibration dataset paths for each model
                - custom_end_node_list: Optional list of custom end_node_names for each model
                - end_node_yaml_path: Optional path to end_node.yaml file (default: "end_node.yaml")
        """
        # Extract model configurations
        self.model_configs = config_dict.get('model_configs', {})
        self.model_names = list(self.model_configs.keys())
        
        # Extract other configurations
        self.root_dir = config_dict.get('root_dir', '')
        self.enable_normalization_script = config_dict.get('enable_normalization_script', True)
        self.custom_calib_paths = config_dict.get('custom_calib_paths', None)
        self.custom_end_node_list = config_dict.get('custom_end_node_list', None)
        self.end_node_yaml_path = config_dict.get('end_node_yaml_path', 'end_node.yaml')
        
        # Validate custom_calib_paths if provided
        if self.custom_calib_paths is not None:
            if len(self.custom_calib_paths) != len(self.model_names):
                raise ValueError("custom_calib_paths length must match number of models")
        
        # Validate custom_end_node_list if provided
        if self.custom_end_node_list is not None:
            if len(self.custom_end_node_list) != len(self.model_names):
                raise ValueError("custom_end_node_list length must match number of models")
        
        # Initialize internal attributes
        self.convert_table = self._load_convert_table()
        self.end_node_names_list = self._build_end_node_names_list()
        self.onnx_paths = []
        self.calib_paths = []
        self.runners = []
        self.r_main = None
        self.output_folder = f"{self.root_dir}/{'_'.join(self.model_names)}/"


    def _load_convert_table(self) -> Dict:
        """
        Load convert table from YAML file.
        
        Returns:
            Dict: Convert table dictionary loaded from YAML file.
        
        Raises:
            FileNotFoundError: If YAML file is not found and no fallback is available.
        """
        # Load from YAML file
        yaml_config = load_end_node_config(self.end_node_yaml_path)
        if yaml_config is not None:
            return yaml_config
        else:
            # If YAML file is not found, raise an error since we no longer have CONVERT_TABLE fallback
            raise FileNotFoundError(f"YAML configuration file '{self.end_node_yaml_path}' not found. Please ensure the file exists and contains the required model configurations.")
    
    def _build_end_node_names_list(self) -> List[List[str]]:
        """
        Build end_node_names_list from either custom list or YAML configuration.
        
        Returns:
            List[List[str]]: List of end_node_names for each model.
        """
        # If custom_end_node_list is provided, use it directly
        if self.custom_end_node_list is not None:
            return self.custom_end_node_list
        
        # Otherwise, build from YAML configuration
        end_node_names_list = []
        for model_name in self.model_names:
            model_type, task, profile_key = self.model_configs[model_name]
            end_node_names = get_end_node_names_from_dict(self.convert_table, model_type, task, profile_key)
            end_node_names_list.append(end_node_names)
        return end_node_names_list

    def check_file_exists(self, file_path: str):
        """
        Check if a file exists. Raise FileNotFoundError if not found.

        Args:
            file_path (str): Path to check.
        """
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File not found: {file_path}")

    def translate_models(self):
        """
        Translate all ONNX models and create corresponding ClientRunner objects.
        Checks for ONNX and calibration file existence.
        Supports custom calibration dataset paths for each model.
        """
        if len(self.model_names) != len(self.end_node_names_list):
            raise ValueError("model_names and end_node_names_list must have the same length.")
        
        for idx, (name, end_node_names) in enumerate(zip(self.model_names, self.end_node_names_list)):
            onnx_path = f"{self.root_dir}/{name}/{name}.onnx"
            print(f"onnx_path: {onnx_path}")
            
            # Use custom calib path if provided, otherwise use default path
            if self.custom_calib_paths is not None:
                calib_path = self.custom_calib_paths[idx]
            else:
                calib_path = f"{self.root_dir}/{name}/calib_set.npy"
            
            self.check_file_exists(onnx_path)
            self.check_file_exists(calib_path)
            self.onnx_paths.append(onnx_path)
            self.calib_paths.append(calib_path)
            runner = ClientRunner(hw_arch="hailo8l")
            runner.translate_onnx_model(onnx_path, end_node_names=end_node_names)
            self.runners.append(runner)

    def join_models(self):
        """
        Join all runners into a single main runner.
        For the first join, rename both runners' "model" scope to their respective model names.
        For subsequent joins, keep existing scopes and rename the new runner's "model" scope.
        """
        self.r_main = self.runners[0]
        if len(self.runners) > 1:
            # First join: rename both runners' "model" scope
            scope1 = {"model": self.model_names[0]}
            scope2 = {"model": self.model_names[1]}
            self.r_main.join(self.runners[1], scope1_name=scope1, scope2_name=scope2, join_action=JoinAction.NONE)
            
            # Subsequent joins: keep existing scopes, rename new runner's "model" scope
            for idx in range(2, len(self.runners)):
                scope1 = {name: name for name in self.model_names[:idx]}
                scope2 = {"model": self.model_names[idx]}
                self.r_main.join(self.runners[idx], scope1_name=scope1, scope2_name=scope2, join_action=JoinAction.NONE)
        print('----- Join Finished.-----')

    def add_normalization(self):
        """
        Insert normalization script if enabled.
        Format: all norm nodes joined by comma, then assigned to one normalization.
        """
        if not self.enable_normalization_script:
            return
        norm_nodes = []
        for i, name in enumerate(self.model_names, 1):
            norm_nodes.append(f"{name}/norm{i}")
        alls = ", ".join(norm_nodes) + " = normalization([0.0, 0.0, 0.0], [255.0, 255.0, 255.0])"
        self.r_main.load_model_script(alls)

    def add_model_script(self, model_script: str):
        self.r_main.load_model_script(model_script)

    def optimize_full_precision(self):
        """
        Run full precision optimization on the main runner.
        """
        self.r_main.optimize_full_precision()

    def calibrate_and_optimize(self, transform_method: list, min_calib: int = None):
        """
        Load all calibration data and run quantization optimization.
        Uses up to 5000 samples, aligned across all models.
        """

        calibs = [transform(np.load(path)) for path, transform in zip(self.calib_paths, transform_method)]
        if isinstance(min_calib, int):
            min_calib = min_calib
        else:
            min_calib = min(len(c) for c in calibs)
        if len(self.runners) > 1:
            calib_dict = {f"{name}/input_layer1": c[:min_calib] for name, c in zip(self.model_names, calibs)}
            print("calib_dict keys:", calib_dict.keys())
            self.r_main.optimize(calib_dict)
        else:
            self.r_main.optimize(calibs[0])

    def compile_and_save(self):
        """
        Compile the main runner and save the HEF file to output_folder.
        """
        hef = self.r_main.compile()
        os.makedirs(self.output_folder, exist_ok=True)
        hef_filename = f"{'_'.join(self.model_names)}.hef"
        hef_path = os.path.join(self.output_folder, hef_filename)
        with open(hef_path, "wb") as f:
            f.write(hef)
        print(f"HEF file saved to: {hef_path}")

    def run(self, transform_method: list, model_script: str, min_calib: int = None):
        """
        Run the full pipeline: translate, join, normalization, optimize, calibrate, and compile.
        """
        self.translate_models()
        self.join_models()
        self.add_normalization()
        if model_script is not None:
            self.add_model_script(model_script)
        self.optimize_full_precision()
        self.calibrate_and_optimize(transform_method, min_calib)
        self.compile_and_save()







from ultralytics.utils import ops
from ultralytics.nn.autobackend import AutoBackend
import cv2
import time
import yaml
import torch
import torchvision
import numpy as np
from typing import Union
from dataclasses import dataclass

def clip_coords_pose(coords, shape):
    """
    Clip line coordinates to the image boundaries.

    Args:
        coords (torch.Tensor | numpy.ndarray): A list of line coordinates.
        shape (tuple): A tuple of integers representing the size of the image in the format (height, width).

    Returns:
        (torch.Tensor | numpy.ndarray): Clipped coordinates
    """
    if isinstance(coords, torch.Tensor):  # faster individually (WARNING: inplace .clamp_() Apple MPS bug)
        coords[..., 0] = coords[..., 0].clamp(0, shape[1] - 1)  # x
        coords[..., 1] = coords[..., 1].clamp(0, shape[0] - 1)  # y
    else:  # np.array (faster grouped)
        coords[..., 0] = coords[..., 0].clip(0, shape[1] - 1)  # x
        coords[..., 1] = coords[..., 1].clip(0, shape[0] - 1)  # y
        
    return coords


def clip_coords(boxes, img_shape):
    # Clip bounding xyxy bounding boxes to image shape (height, width)
    boxes[:, 0].clamp_(0, img_shape[1] - 1)  # x1
    boxes[:, 1].clamp_(0, img_shape[0] - 1)  # y1
    boxes[:, 2].clamp_(0, img_shape[1] - 1)  # x2
    boxes[:, 3].clamp_(0, img_shape[0] - 1)  # y2

def scale_coords_pose(img1_shape, coords, img0_shape, ratio_pad=None, normalize=False, padding=True):
    """
    Rescale segment coordinates (xy) from img1_shape to img0_shape.

    Args:
        img1_shape (tuple): The shape of the image that the coords are from.
        coords (torch.Tensor): the coords to be scaled of shape n,2.
        img0_shape (tuple): the shape of the image that the segmentation is being applied to.
        ratio_pad (tuple): the ratio of the image size to the padded image size.
        normalize (bool): If True, the coordinates will be normalized to the range [0, 1]. Defaults to False.
        padding (bool): If True, assuming the boxes is based on image augmented by yolo style. If False then do regular
            rescaling.

    Returns:
        coords (torch.Tensor): The scaled coordinates.
    """
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    if padding:
        coords[..., 0] -= pad[0]  # x padding
        coords[..., 1] -= pad[1]  # y padding
    coords[..., 0] /= gain
    coords[..., 1] /= gain
    coords = clip_coords_pose(coords, img0_shape)
    if normalize:
        coords[..., 0] /= img0_shape[1]  # width
        coords[..., 1] /= img0_shape[0]  # height
        
    return coords


def scale_coords(img1_shape, coords, img0_shape, ratio_pad=None):
    # Rescale coords (xyxy) from img1_shape to img0_shape
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    coords[:, [0, 2]] -= pad[0]  # x padding
    coords[:, [1, 3]] -= pad[1]  # y padding
    coords[:, :4] /= gain
    clip_coords(coords, img0_shape)
    return coords


def letterbox(img, new_shape=(640, 640), color=(114, 114, 114), auto=True, scaleFill=False, scaleup=True):
    # Resize image to a 32-pixel-multiple rectangle https://github.com/ultralytics/yolov3/issues/232
    shape = img.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better test mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, 64), np.mod(dh, 64)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = (new_shape[1], new_shape[0])
        ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

    dw /= 2  # divide padding into 2 sides
    dh /= 2
    
    if shape[::-1] != new_unpad:  # resize
        img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    
    return img, ratio, (dw, dh)

class TorchPoseEstimator:
    def __init__(
        self, 
        model, 
        device, 
        size,
        draw_settings,
        conf_thres=0.25,
        iou_thres=0.7,
        fp16=False 
    ):
        """
        初始化姿勢估計器。

        :param model: 用於姿勢估計的模型。
        :param device: 執行模型的設備（例如 'cpu' 或 'cuda'）。
        :param size: 輸入圖像的尺寸。
        :param conf_thres: 檢測的置信度閾值。
        :param iou_thres: 非最大抑制的交叉聯合比閾值。
        :param fp16: 是否使用半精度浮點數（對於某些GPU加速有用）。
        """
        self.model = model.half().eval() if fp16 else model.float().eval()
        self.device = device
        self.size = size
        self.conf_thres = conf_thres
        self.iou_thres = iou_thres
        self.fps16 = fp16
        self.draw_settings = draw_settings
        self._initialize_draw_settings()
        
    def _initialize_draw_settings(self):
        with open(self.draw_settings, 'r') as f:
            draw_settings = yaml.safe_load(f)
        self.bbox_color = tuple(draw_settings['bbox_color'])
        self.bbox_thickness = draw_settings['bbox_thickness']
        self.bbox_labelstr = draw_settings['bbox_labelstr']
        self.kpt_color_map = draw_settings['kpt_color_map']
        self.kpt_labelstr = draw_settings['kpt_labelstr']
        self.skeleton_map = draw_settings['skeleton_map']
        self.kpt_labelstr = draw_settings['kpt_labelstr']

    def __call__(self, imgs):
        """
        對一個或多個圖像進行姿勢估計。

        :param imgs: 圖像數據，可以是單一圖像或圖像列表。
        :return: 每個圖像的姿勢估計結果。
        """
        if not len(imgs):
            return [] 
        
        # preprocess
        if isinstance(imgs, np.ndarray):
            imgs = [imgs] 
            
        pre_imgs = []
        for img in imgs:
            pre_img = letterbox(img, self.size, auto=False)[0]
            pre_img = pre_img.transpose((2, 0, 1))[::-1]  # HWC -> CHW, BGR -> RGB
            pre_img = np.ascontiguousarray(pre_img)
            pre_img = torch.from_numpy(pre_img).float().to(self.device)
            pre_img /= 255
            if self.fps16:
                pre_img = pre_img.half()
            pre_imgs.append(pre_img)
        pre_imgs = torch.stack(pre_imgs) # 將列表轉為tensor
        
        with torch.no_grad():
            
            preds = self.model(pre_imgs)
            preds = ops.non_max_suppression(
                preds,
                conf_thres=self.conf_thres,
                iou_thres=self.iou_thres,
                nc=1, # 代表只有一個類別，也就是人
            )
        results = []
        for i, det in enumerate(preds):
            if len(det):
                det[:, :4] = scale_coords(pre_imgs.shape[2:], det[:, :4], imgs[i].shape).round()
                pred_det = det[:, :6].cpu().numpy() # [x1, y1, x2, y2, conf, cls]
                bbox_xyxy = pred_det[:, :4].astype(int)
                
                pred_kpts = det[:, 6:].view(len(det), self.model.kpt_shape[0], self.model.kpt_shape[1]) # [n, 17, 2]
                pred_kpts = scale_coords_pose(pre_imgs.shape[2:], pred_kpts, imgs[i].shape)
                pred_kpts = pred_kpts.cpu().numpy()
                pred_kpts[..., :2] = pred_kpts[..., :2].astype(int)
                bbox_keypoints = pred_kpts
            else:
                bbox_xyxy = np.zeros((0, 4), dtype=int) # 因為沒有偵測到人，所以bbox為空
                bbox_keypoints = np.zeros((0, self.model.kpt_shape[0], self.model.kpt_shape[1]), dtype=int) # 因為沒有偵測到人，所以keypoints為空
            images_result = {
                'bbox': bbox_xyxy,
                'keypoints': bbox_keypoints
            }
            results.append(images_result)
            
        return results
    
    def draw_skeleton_on_image(self, img, bbox_keypoints, conf_thres=0.25):
        for skeleton in self.skeleton_map:
            srt_kpt_id = skeleton['srt_kpt_id']
            srt_kpt_x = bbox_keypoints[srt_kpt_id][0]
            srt_kpt_y = bbox_keypoints[srt_kpt_id][1]
            srt_kpt_score = bbox_keypoints[srt_kpt_id][2]
            if srt_kpt_score < conf_thres:
                continue

            if srt_kpt_x == 0 and srt_kpt_y == 0:
                continue

            dst_kpt_id = skeleton['dst_kpt_id']
            dst_kpt_x = bbox_keypoints[dst_kpt_id][0]
            dst_kpt_y = bbox_keypoints[dst_kpt_id][1]
            dst_kpt_score = bbox_keypoints[dst_kpt_id][2]
            if dst_kpt_score < conf_thres:
                continue

            if dst_kpt_x == 0 and dst_kpt_y == 0:
                continue

            skeleton_color = skeleton['color']
            skeleton_thickness = skeleton['thickness']

            img = cv2.line(
                img,
                (int(srt_kpt_x), int(srt_kpt_y)),
                (int(dst_kpt_x), int(dst_kpt_y)),
                color=skeleton_color,
                thickness=skeleton_thickness
            )

        return img
    
    def draw_keypoints_on_image(self, img, bbox_keypoints, show_id=False, conf_thres=0.25):
        for kpt_id in self.kpt_color_map:
            kpt_color_info = self.kpt_color_map[kpt_id]
            kpt_color = kpt_color_info['color']
            kpt_radius = kpt_color_info['radius']
            kpt_x = bbox_keypoints[kpt_id][0]
            kpt_y = bbox_keypoints[kpt_id][1]
            kpt_score = bbox_keypoints[kpt_id][2]

            if kpt_score < conf_thres:
                continue

            img = cv2.circle(
                img,
                (int(kpt_x), int(kpt_y)),
                radius=kpt_radius,
                color=kpt_color,
                thickness=-1
            )

            if show_id:
                kpt_label = str(kpt_id)
                offset_x = self.kpt_labelstr.get('offset_x', 0)
                offset_y = self.kpt_labelstr.get('offset_y', 0)
                font_size = self.kpt_labelstr.get('font_size', 0.5)
                font_thickness = self.kpt_labelstr.get('font_thickness', 1)

                img = cv2.putText(
                    img,
                    kpt_label,
                    (int(kpt_x + offset_x), int(kpt_y + offset_y)),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    font_size,
                    kpt_color,
                    font_thickness
                )

        return img 

    def draw_pose_results_on_images(self, image, bboxes, keypoints, show_id=False, conf_thres=0.25): 
        min_length = min(len(bboxes), len(keypoints))
        bboxes = bboxes[:min_length]
        keypoints = keypoints[:min_length]
        for bbox, kpts in zip(bboxes, keypoints):
     
            if bbox.size != 0: 
                cv2.rectangle(
                    image,
                    (int(bbox[0]), int(bbox[1])),
                    (int(bbox[2]), int(bbox[3])),
                    color=self.bbox_color,
                    thickness=self.bbox_thickness
                )

      
            if len(kpts) != 0:  
                image = self.draw_keypoints_on_image(image, kpts, show_id, conf_thres)
                image = self.draw_skeleton_on_image(image, kpts, conf_thres)
        
        return image
    
    def destroy(self):
        if self.model:
            del self.model
    
# =============================================================================
# PoseEstimatorBuilder
# ============================================================================= 

@dataclass(frozen=True)
class PoseEstimatorOptions:
    weights: str = './weights/yolov8x-pose.pt'
    draw_settings: str = './configuration/body_pose_settings.yaml'
    size: Union[int, tuple] = 640
    conf_threshold: float = 0.25
    iou_threshold: float = 0.7
    fp16: bool = False
    verbose: bool = False

def PoseEstimator(config: Union[str, PoseEstimatorOptions]):
    """
    A class to initializes and load the PoseEstimator model based on a YAML configuration or an object.

    Args:
            config (Union[str, PoseEstimatorOptions]): A YAML configuration file path or a PoseEstimatorOptions object.
        Raises:
            ValueError: If the config isn't a YAML file or 'PoseEstimatorOptions' object.

        Example:
            1. YAML file
                >>> model = PoseEstimator("config.yaml")
                >>> results = model(images)
                >>> print(results)
            2. object
                >>> options = PoseEstimatorOptions(conf_threshold=0.6, verbose=True....)
                >>> model = PoseEstimator(options)
                >>> results = model(images)
                >>> print(results)
        """
    
    # Load config from a YAML file or a PoseEstimatorOptions object
    if isinstance(config, str) and config.endswith(('yaml', 'yml')):
        with open(config, 'r') as f:
            setting = yaml.safe_load(f)

    elif isinstance(config, PoseEstimatorOptions):
        setting = {k: v for k, v in config.__dict__.items() if not k.startswith("__")}

    else:
        raise ValueError("Config must be a YAML file or an object.")
    
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    # Load model
    model = AutoBackend(
        weights=setting['weights'],
        device=device,
        fp16=setting.get('fp16', False),
        verbose=setting.get('verbose', False)
    )

    # Load the pose estimator model based on file type
    pose_estimator = TorchPoseEstimator(
        model=model,
        device=device.type,  # 'cuda' or 'cpu'
        size=setting.get('size', 640), 
        draw_settings=setting['draw_settings'],
        conf_thres=setting.get('conf_threshold', 0.25), 
        iou_thres=setting.get('iou_threshold', 0.7),
        fp16=setting.get('fp16', False)
    )

    return pose_estimator



import os

def is_jetson():
    """Check if the code is running on an NVIDIA Jetson platform."""
    compat_file_path = "/proc/device-tree/compatible"
    if os.path.isfile(compat_file_path):
        with open(compat_file_path, "r") as file:
            compat_str = file.read()
            return "nvidia" in compat_str and "tegra" in compat_str
    return False


# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
General utils
"""

import contextlib
import glob
import inspect
import logging
import math
import os
import platform
import random
import re
import shutil
import signal
import sys
import time
import urllib
from copy import deepcopy
from datetime import datetime
from itertools import repeat
from multiprocessing.pool import ThreadPool
from pathlib import Path
from subprocess import check_output
from typing import Optional
from zipfile import ZipFile

import cv2
import numpy as np
import pandas as pd
import pkg_resources as pkg
import torch
import torchvision
import yaml

from utils import TryExcept, emojis
from utils.downloads import gsutil_getsize
from utils.metrics import box_iou, fitness

FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]  # YOLOv5 root directory
RANK = int(os.getenv('RANK', -1))

# Settings
NUM_THREADS = min(8, max(1, os.cpu_count() - 1))  # number of YOLOv5 multiprocessing threads
DATASETS_DIR = Path(os.getenv('YOLOv5_DATASETS_DIR', ROOT.parent / 'datasets'))  # global datasets directory
AUTOINSTALL = str(os.getenv('YOLOv5_AUTOINSTALL', True)).lower() == 'true'  # global auto-install mode
VERBOSE = str(os.getenv('YOLOv5_VERBOSE', True)).lower() == 'true'  # global verbose mode
FONT = 'Arial.ttf'  # https://ultralytics.com/assets/Arial.ttf

torch.set_printoptions(linewidth=320, precision=5, profile='long')
np.set_printoptions(linewidth=320, formatter={'float_kind': '{:11.5g}'.format})  # format short g, %precision=5
pd.options.display.max_columns = 10
cv2.setNumThreads(0)  # prevent OpenCV from multithreading (incompatible with PyTorch DataLoader)
os.environ['NUMEXPR_MAX_THREADS'] = str(NUM_THREADS)  # NumExpr max threads
os.environ['OMP_NUM_THREADS'] = '1' if platform.system() == 'darwin' else str(NUM_THREADS)  # OpenMP (PyTorch and SciPy)


def is_ascii(s=''):
    # Is string composed of all ASCII (no UTF) characters? (note str().isascii() introduced in python 3.7)
    s = str(s)  # convert list, tuple, None, etc. to str
    return len(s.encode().decode('ascii', 'ignore')) == len(s)


def is_chinese(s='人工智能'):
    # Is string composed of any Chinese characters?
    return bool(re.search('[\u4e00-\u9fff]', str(s)))


def is_colab():
    # Is environment a Google Colab instance?
    return 'COLAB_GPU' in os.environ


def is_kaggle():
    # Is environment a Kaggle Notebook?
    return os.environ.get('PWD') == '/kaggle/working' and os.environ.get('KAGGLE_URL_BASE') == 'https://www.kaggle.com'


def is_docker() -> bool:
    """Check if the process runs inside a docker container."""
    if Path("/.dockerenv").exists():
        return True
    try:  # check if docker is in control groups
        with open("/proc/self/cgroup") as file:
            return any("docker" in line for line in file)
    except OSError:
        return False


def is_writeable(dir, test=False):
    # Return True if directory has write permissions, test opening a file with write permissions if test=True
    if not test:
        return os.access(dir, os.W_OK)  # possible issues on Windows
    file = Path(dir) / 'tmp.txt'
    try:
        with open(file, 'w'):  # open file with write permissions
            pass
        file.unlink()  # remove file
        return True
    except OSError:
        return False


def set_logging(name=None, verbose=VERBOSE):
    # Sets level and returns logger
    if is_kaggle() or is_colab():
        for h in logging.root.handlers:
            logging.root.removeHandler(h)  # remove all handlers associated with the root logger object
    rank = int(os.getenv('RANK', -1))  # rank in world for Multi-GPU trainings
    level = logging.INFO if verbose and rank in {-1, 0} else logging.ERROR
    log = logging.getLogger(name)
    log.setLevel(level)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(message)s"))
    handler.setLevel(level)
    log.addHandler(handler)


set_logging()  # run before defining LOGGER
LOGGER = logging.getLogger("yolov5")  # define globally (used in train.py, val.py, detect.py, etc.)
if platform.system() == 'Windows':
    for fn in LOGGER.info, LOGGER.warning:
        setattr(LOGGER, fn.__name__, lambda x: fn(emojis(x)))  # emoji safe logging


def user_config_dir(dir='Ultralytics', env_var='YOLOV5_CONFIG_DIR'):
    # Return path of user configuration directory. Prefer environment variable if exists. Make dir if required.
    env = os.getenv(env_var)
    if env:
        path = Path(env)  # use environment variable
    else:
        cfg = {'Windows': 'AppData/Roaming', 'Linux': '.config', 'Darwin': 'Library/Application Support'}  # 3 OS dirs
        path = Path.home() / cfg.get(platform.system(), '')  # OS-specific config dir
        path = (path if is_writeable(path) else Path('/tmp')) / dir  # GCP and AWS lambda fix, only /tmp is writeable
    path.mkdir(exist_ok=True)  # make if required
    return path


CONFIG_DIR = user_config_dir()  # Ultralytics settings dir


class Profile(contextlib.ContextDecorator):
    # YOLOv5 Profile class. Usage: @Profile() decorator or 'with Profile():' context manager
    def __init__(self, t=0.0):
        self.t = t
        self.cuda = torch.cuda.is_available()

    def __enter__(self):
        self.start = self.time()
        return self

    def __exit__(self, type, value, traceback):
        self.dt = self.time() - self.start  # delta-time
        self.t += self.dt  # accumulate dt

    def time(self):
        if self.cuda:
            torch.cuda.synchronize()
        return time.time()


class Timeout(contextlib.ContextDecorator):
    # YOLOv5 Timeout class. Usage: @Timeout(seconds) decorator or 'with Timeout(seconds):' context manager
    def __init__(self, seconds, *, timeout_msg='', suppress_timeout_errors=True):
        self.seconds = int(seconds)
        self.timeout_message = timeout_msg
        self.suppress = bool(suppress_timeout_errors)

    def _timeout_handler(self, signum, frame):
        raise TimeoutError(self.timeout_message)

    def __enter__(self):
        if platform.system() != 'Windows':  # not supported on Windows
            signal.signal(signal.SIGALRM, self._timeout_handler)  # Set handler for SIGALRM
            signal.alarm(self.seconds)  # start countdown for SIGALRM to be raised

    def __exit__(self, exc_type, exc_val, exc_tb):
        if platform.system() != 'Windows':
            signal.alarm(0)  # Cancel SIGALRM if it's scheduled
            if self.suppress and exc_type is TimeoutError:  # Suppress TimeoutError
                return True


class WorkingDirectory(contextlib.ContextDecorator):
    # Usage: @WorkingDirectory(dir) decorator or 'with WorkingDirectory(dir):' context manager
    def __init__(self, new_dir):
        self.dir = new_dir  # new dir
        self.cwd = Path.cwd().resolve()  # current dir

    def __enter__(self):
        os.chdir(self.dir)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.cwd)


def methods(instance):
    # Get class/instance methods
    return [f for f in dir(instance) if callable(getattr(instance, f)) and not f.startswith("__")]


def print_args(args: Optional[dict] = None, show_file=True, show_func=False):
    # Print function arguments (optional args dict)
    x = inspect.currentframe().f_back  # previous frame
    file, _, func, _, _ = inspect.getframeinfo(x)
    if args is None:  # get args automatically
        args, _, _, frm = inspect.getargvalues(x)
        args = {k: v for k, v in frm.items() if k in args}
    try:
        file = Path(file).resolve().relative_to(ROOT).with_suffix('')
    except ValueError:
        file = Path(file).stem
    s = (f'{file}: ' if show_file else '') + (f'{func}: ' if show_func else '')
    LOGGER.info(colorstr(s) + ', '.join(f'{k}={v}' for k, v in args.items()))


def init_seeds(seed=0, deterministic=False):
    # Initialize random number generator (RNG) seeds https://pytorch.org/docs/stable/notes/randomness.html
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)  # for Multi-GPU, exception safe
    # torch.backends.cudnn.benchmark = True  # AutoBatch problem https://github.com/ultralytics/yolov5/issues/9287
    if deterministic and check_version(torch.__version__, '1.12.0'):  # https://github.com/ultralytics/yolov5/pull/8213
        torch.use_deterministic_algorithms(True)
        torch.backends.cudnn.deterministic = True
        os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8'
        os.environ['PYTHONHASHSEED'] = str(seed)


def intersect_dicts(da, db, exclude=()):
    # Dictionary intersection of matching keys and shapes, omitting 'exclude' keys, using da values
    return {k: v for k, v in da.items() if k in db and all(x not in k for x in exclude) and v.shape == db[k].shape}


def get_default_args(func):
    # Get func() default arguments
    signature = inspect.signature(func)
    return {k: v.default for k, v in signature.parameters.items() if v.default is not inspect.Parameter.empty}


def get_latest_run(search_dir='.'):
    # Return path to most recent 'last.pt' in /runs (i.e. to --resume from)
    last_list = glob.glob(f'{search_dir}/**/last*.pt', recursive=True)
    return max(last_list, key=os.path.getctime) if last_list else ''


def file_age(path=__file__):
    # Return days since last file update
    dt = (datetime.now() - datetime.fromtimestamp(Path(path).stat().st_mtime))  # delta
    return dt.days  # + dt.seconds / 86400  # fractional days


def file_date(path=__file__):
    # Return human-readable file modification date, i.e. '2021-3-26'
    t = datetime.fromtimestamp(Path(path).stat().st_mtime)
    return f'{t.year}-{t.month}-{t.day}'


def file_size(path):
    # Return file/dir size (MB)
    mb = 1 << 20  # bytes to MiB (1024 ** 2)
    path = Path(path)
    if path.is_file():
        return path.stat().st_size / mb
    elif path.is_dir():
        return sum(f.stat().st_size for f in path.glob('**/*') if f.is_file()) / mb
    else:
        return 0.0


def check_online():
    # Check internet connectivity
    import socket
    try:
        socket.create_connection(("1.1.1.1", 443), 5)  # check host accessibility
        return True
    except OSError:
        return False


def git_describe(path=ROOT):  # path must be a directory
    # Return human-readable git description, i.e. v5.0-5-g3e25f1e https://git-scm.com/docs/git-describe
    try:
        assert (Path(path) / '.git').is_dir()
        return check_output(f'git -C {path} describe --tags --long --always', shell=True).decode()[:-1]
    except Exception:
        return ''


@TryExcept()
@WorkingDirectory(ROOT)
def check_git_status(repo='ultralytics/yolov5', branch='master'):
    # YOLOv5 status check, recommend 'git pull' if code is out of date
    url = f'https://github.com/{repo}'
    msg = f', for updates see {url}'
    s = colorstr('github: ')  # string
    assert Path('.git').exists(), s + 'skipping check (not a git repository)' + msg
    assert check_online(), s + 'skipping check (offline)' + msg

    splits = re.split(pattern=r'\s', string=check_output('git remote -v', shell=True).decode())
    matches = [repo in s for s in splits]
    if any(matches):
        remote = splits[matches.index(True) - 1]
    else:
        remote = 'ultralytics'
        check_output(f'git remote add {remote} {url}', shell=True)
    check_output(f'git fetch {remote}', shell=True, timeout=5)  # git fetch
    local_branch = check_output('git rev-parse --abbrev-ref HEAD', shell=True).decode().strip()  # checked out
    n = int(check_output(f'git rev-list {local_branch}..{remote}/{branch} --count', shell=True))  # commits behind
    if n > 0:
        pull = 'git pull' if remote == 'origin' else f'git pull {remote} {branch}'
        s += f"⚠️ YOLOv5 is out of date by {n} commit{'s' * (n > 1)}. Use `{pull}` or `git clone {url}` to update."
    else:
        s += f'up to date with {url} ✅'
    LOGGER.info(s)


def check_python(minimum='3.7.0'):
    # Check current python version vs. required python version
    check_version(platform.python_version(), minimum, name='Python ', hard=True)


def check_version(current='0.0.0', minimum='0.0.0', name='version ', pinned=False, hard=False, verbose=False):
    # Check version vs. required version
    current, minimum = (pkg.parse_version(x) for x in (current, minimum))
    result = (current == minimum) if pinned else (current >= minimum)  # bool
    s = f'WARNING ⚠️ {name}{minimum} is required by YOLOv5, but {name}{current} is currently installed'  # string
    if hard:
        assert result, emojis(s)  # assert min requirements met
    if verbose and not result:
        LOGGER.warning(s)
    return result


@TryExcept()
def check_requirements(requirements=ROOT / 'requirements.txt', exclude=(), install=True, cmds=''):
    # Check installed dependencies meet YOLOv5 requirements (pass *.txt file or list of packages or single package str)
    prefix = colorstr('red', 'bold', 'requirements:')
    check_python()  # check python version
    if isinstance(requirements, Path):  # requirements.txt file
        file = requirements.resolve()
        assert file.exists(), f"{prefix} {file} not found, check failed."
        with file.open() as f:
            requirements = [f'{x.name}{x.specifier}' for x in pkg.parse_requirements(f) if x.name not in exclude]
    elif isinstance(requirements, str):
        requirements = [requirements]

    s = ''
    n = 0
    for r in requirements:
        try:
            pkg.require(r)
        except (pkg.VersionConflict, pkg.DistributionNotFound):  # exception if requirements not met
            s += f'"{r}" '
            n += 1

    if s and install and AUTOINSTALL:  # check environment variable
        LOGGER.info(f"{prefix} YOLOv5 requirement{'s' * (n > 1)} {s}not found, attempting AutoUpdate...")
        try:
            assert check_online(), "AutoUpdate skipped (offline)"
            LOGGER.info(check_output(f'pip install {s} {cmds}', shell=True).decode())
            source = file if 'file' in locals() else requirements
            s = f"{prefix} {n} package{'s' * (n > 1)} updated per {source}\n" \
                f"{prefix} ⚠️ {colorstr('bold', 'Restart runtime or rerun command for updates to take effect')}\n"
            LOGGER.info(s)
        except Exception as e:
            LOGGER.warning(f'{prefix} ❌ {e}')


def check_img_size(imgsz, s=32, floor=0):
    # Verify image size is a multiple of stride s in each dimension
    if isinstance(imgsz, int):  # integer i.e. img_size=640
        new_size = max(make_divisible(imgsz, int(s)), floor)
    else:  # list i.e. img_size=[640, 480]
        imgsz = list(imgsz)  # convert to list if tuple
        new_size = [max(make_divisible(x, int(s)), floor) for x in imgsz]
    if new_size != imgsz:
        LOGGER.warning(f'WARNING ⚠️ --img-size {imgsz} must be multiple of max stride {s}, updating to {new_size}')
    return new_size


def check_imshow():
    # Check if environment supports image displays
    try:
        assert not is_docker(), 'cv2.imshow() is disabled in Docker environments'
        assert not is_colab(), 'cv2.imshow() is disabled in Google Colab environments'
        cv2.imshow('test', np.zeros((1, 1, 3)))
        cv2.waitKey(1)
        cv2.destroyAllWindows()
        cv2.waitKey(1)
        return True
    except Exception as e:
        LOGGER.warning(f'WARNING ⚠️ Environment does not support cv2.imshow() or PIL Image.show() image displays\n{e}')
        return False


def check_suffix(file='yolov5s.pt', suffix=('.pt',), msg=''):
    # Check file(s) for acceptable suffix
    if file and suffix:
        if isinstance(suffix, str):
            suffix = [suffix]
        for f in file if isinstance(file, (list, tuple)) else [file]:
            s = Path(f).suffix.lower()  # file suffix
            if len(s):
                assert s in suffix, f"{msg}{f} acceptable suffix is {suffix}"


def check_yaml(file, suffix=('.yaml', '.yml')):
    # Search/download YAML file (if necessary) and return path, checking suffix
    return check_file(file, suffix)


def check_file(file, suffix=''):
    # Search/download file (if necessary) and return path
    check_suffix(file, suffix)  # optional
    file = str(file)  # convert to str()
    if Path(file).is_file() or not file:  # exists
        return file
    elif file.startswith(('http:/', 'https:/')):  # download
        url = file  # warning: Pathlib turns :// -> :/
        file = Path(urllib.parse.unquote(file).split('?')[0]).name  # '%2F' to '/', split https://url.com/file.txt?auth
        if Path(file).is_file():
            LOGGER.info(f'Found {url} locally at {file}')  # file already exists
        else:
            LOGGER.info(f'Downloading {url} to {file}...')
            torch.hub.download_url_to_file(url, file)
            assert Path(file).exists() and Path(file).stat().st_size > 0, f'File download failed: {url}'  # check
        return file
    elif file.startswith('clearml://'):  # ClearML Dataset ID
        assert 'clearml' in sys.modules, "ClearML is not installed, so cannot use ClearML dataset. Try running 'pip install clearml'."
        return file
    else:  # search
        files = []
        for d in 'data', 'models', 'utils':  # search directories
            files.extend(glob.glob(str(ROOT / d / '**' / file), recursive=True))  # find file
        assert len(files), f'File not found: {file}'  # assert file was found
        assert len(files) == 1, f"Multiple files match '{file}', specify exact path: {files}"  # assert unique
        return files[0]  # return file


def check_font(font=FONT, progress=False):
    # Download font to CONFIG_DIR if necessary
    font = Path(font)
    file = CONFIG_DIR / font.name
    if not font.exists() and not file.exists():
        url = f'https://ultralytics.com/assets/{font.name}'
        LOGGER.info(f'Downloading {url} to {file}...')
        torch.hub.download_url_to_file(url, str(file), progress=progress)


def check_dataset(data, autodownload=True):
    # Download, check and/or unzip dataset if not found locally

    # Download (optional)
    extract_dir = ''
    if isinstance(data, (str, Path)) and str(data).endswith('.zip'):  # i.e. gs://bucket/dir/coco128.zip
        download(data, dir=f'{DATASETS_DIR}/{Path(data).stem}', unzip=True, delete=False, curl=False, threads=1)
        data = next((DATASETS_DIR / Path(data).stem).rglob('*.yaml'))
        extract_dir, autodownload = data.parent, False

    # Read yaml (optional)
    if isinstance(data, (str, Path)):
        data = yaml_load(data)  # dictionary

    # Checks
    for k in 'train', 'val', 'names':
        assert k in data, f"data.yaml '{k}:' field missing ❌"
    if isinstance(data['names'], (list, tuple)):  # old array format
        data['names'] = dict(enumerate(data['names']))  # convert to dict
    data['nc'] = len(data['names'])

    # Resolve paths
    path = Path(extract_dir or data.get('path') or '')  # optional 'path' default to '.'
    if not path.is_absolute():
        path = (ROOT / path).resolve()
    for k in 'train', 'val', 'test':
        if data.get(k):  # prepend path
            if isinstance(data[k], str):
                x = (path / data[k]).resolve()
                if not x.exists() and data[k].startswith('../'):
                    x = (path / data[k][3:]).resolve()
                data[k] = str(x)
            else:
                data[k] = [str((path / x).resolve()) for x in data[k]]

    # Parse yaml
    train, val, test, s = (data.get(x) for x in ('train', 'val', 'test', 'download'))
    if val:
        val = [Path(x).resolve() for x in (val if isinstance(val, list) else [val])]  # val path
        if not all(x.exists() for x in val):
            LOGGER.info('\nDataset not found ⚠️, missing paths %s' % [str(x) for x in val if not x.exists()])
            if not s or not autodownload:
                raise Exception('Dataset not found ❌')
            t = time.time()
            if s.startswith('http') and s.endswith('.zip'):  # URL
                f = Path(s).name  # filename
                LOGGER.info(f'Downloading {s} to {f}...')
                torch.hub.download_url_to_file(s, f)
                Path(DATASETS_DIR).mkdir(parents=True, exist_ok=True)  # create root
                ZipFile(f).extractall(path=DATASETS_DIR)  # unzip
                Path(f).unlink()  # remove zip
                r = None  # success
            elif s.startswith('bash '):  # bash script
                LOGGER.info(f'Running {s} ...')
                r = os.system(s)
            else:  # python script
                r = exec(s, {'yaml': data})  # return None
            dt = f'({round(time.time() - t, 1)}s)'
            s = f"success ✅ {dt}, saved to {colorstr('bold', DATASETS_DIR)}" if r in (0, None) else f"failure {dt} ❌"
            LOGGER.info(f"Dataset download {s}")
    check_font('Arial.ttf' if is_ascii(data['names']) else 'Arial.Unicode.ttf', progress=True)  # download fonts
    return data  # dictionary


def check_amp(model):
    # Check PyTorch Automatic Mixed Precision (AMP) functionality. Return True on correct operation
    from models.common import AutoShape, DetectMultiBackend

    def amp_allclose(model, im):
        # All close FP32 vs AMP results
        m = AutoShape(model, verbose=False)  # model
        a = m(im).xywhn[0]  # FP32 inference
        m.amp = True
        b = m(im).xywhn[0]  # AMP inference
        return a.shape == b.shape and torch.allclose(a, b, atol=0.1)  # close to 10% absolute tolerance

    prefix = colorstr('AMP: ')
    device = next(model.parameters()).device  # get model device
    if device.type in ('cpu', 'mps'):
        return False  # AMP only used on CUDA devices
    f = ROOT / 'data' / 'images' / 'bus.jpg'  # image to check
    im = f if f.exists() else 'https://ultralytics.com/images/bus.jpg' if check_online() else np.ones((640, 640, 3))
    try:
        assert amp_allclose(deepcopy(model), im) or amp_allclose(DetectMultiBackend('yolov5n.pt', device), im)
        LOGGER.info(f'{prefix}checks passed ✅')
        return True
    except Exception:
        help_url = 'https://github.com/ultralytics/yolov5/issues/7908'
        LOGGER.warning(f'{prefix}checks failed ❌, disabling Automatic Mixed Precision. See {help_url}')
        return False


def yaml_load(file='data.yaml'):
    # Single-line safe yaml loading
    with open(file, errors='ignore') as f:
        return yaml.safe_load(f)


def yaml_save(file='data.yaml', data={}):
    # Single-line safe yaml saving
    with open(file, 'w') as f:
        yaml.safe_dump({k: str(v) if isinstance(v, Path) else v for k, v in data.items()}, f, sort_keys=False)


def url2file(url):
    # Convert URL to filename, i.e. https://url.com/file.txt?auth -> file.txt
    url = str(Path(url)).replace(':/', '://')  # Pathlib turns :// -> :/
    return Path(urllib.parse.unquote(url)).name.split('?')[0]  # '%2F' to '/', split https://url.com/file.txt?auth


def download(url, dir='.', unzip=True, delete=True, curl=False, threads=1, retry=3):
    # Multithreaded file download and unzip function, used in data.yaml for autodownload
    def download_one(url, dir):
        # Download 1 file
        success = True
        if Path(url).is_file():
            f = Path(url)  # filename
        else:  # does not exist
            f = dir / Path(url).name
            LOGGER.info(f'Downloading {url} to {f}...')
            for i in range(retry + 1):
                if curl:
                    s = 'sS' if threads > 1 else ''  # silent
                    r = os.system(
                        f'curl -# -{s}L "{url}" -o "{f}" --retry 9 -C -')  # curl download with retry, continue
                    success = r == 0
                else:
                    torch.hub.download_url_to_file(url, f, progress=threads == 1)  # torch download
                    success = f.is_file()
                if success:
                    break
                elif i < retry:
                    LOGGER.warning(f'⚠️ Download failure, retrying {i + 1}/{retry} {url}...')
                else:
                    LOGGER.warning(f'❌ Failed to download {url}...')

        if unzip and success and f.suffix in ('.zip', '.tar', '.gz'):
            LOGGER.info(f'Unzipping {f}...')
            if f.suffix == '.zip':
                ZipFile(f).extractall(path=dir)  # unzip
            elif f.suffix == '.tar':
                os.system(f'tar xf {f} --directory {f.parent}')  # unzip
            elif f.suffix == '.gz':
                os.system(f'tar xfz {f} --directory {f.parent}')  # unzip
            if delete:
                f.unlink()  # remove zip

    dir = Path(dir)
    dir.mkdir(parents=True, exist_ok=True)  # make directory
    if threads > 1:
        pool = ThreadPool(threads)
        pool.imap(lambda x: download_one(*x), zip(url, repeat(dir)))  # multithreaded
        pool.close()
        pool.join()
    else:
        for u in [url] if isinstance(url, (str, Path)) else url:
            download_one(u, dir)


def make_divisible(x, divisor):
    # Returns nearest x divisible by divisor
    if isinstance(divisor, torch.Tensor):
        divisor = int(divisor.max())  # to int
    return math.ceil(x / divisor) * divisor


def clean_str(s):
    # Cleans a string by replacing special characters with underscore _
    return re.sub(pattern="[|@#!¡·$€%&()=?¿^*;:,¨´><+]", repl="_", string=s)


def one_cycle(y1=0.0, y2=1.0, steps=100):
    # lambda function for sinusoidal ramp from y1 to y2 https://arxiv.org/pdf/1812.01187.pdf
    return lambda x: ((1 - math.cos(x * math.pi / steps)) / 2) * (y2 - y1) + y1


def colorstr(*input):
    # Colors a string https://en.wikipedia.org/wiki/ANSI_escape_code, i.e.  colorstr('blue', 'hello world')
    *args, string = input if len(input) > 1 else ('blue', 'bold', input[0])  # color arguments, string
    colors = {
        'black': '\033[30m',  # basic colors
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'magenta': '\033[35m',
        'cyan': '\033[36m',
        'white': '\033[37m',
        'bright_black': '\033[90m',  # bright colors
        'bright_red': '\033[91m',
        'bright_green': '\033[92m',
        'bright_yellow': '\033[93m',
        'bright_blue': '\033[94m',
        'bright_magenta': '\033[95m',
        'bright_cyan': '\033[96m',
        'bright_white': '\033[97m',
        'end': '\033[0m',  # misc
        'bold': '\033[1m',
        'underline': '\033[4m'}
    return ''.join(colors[x] for x in args) + f'{string}' + colors['end']


def labels_to_class_weights(labels, nc=80):
    # Get class weights (inverse frequency) from training labels
    if labels[0] is None:  # no labels loaded
        return torch.Tensor()

    labels = np.concatenate(labels, 0)  # labels.shape = (866643, 5) for COCO
    classes = labels[:, 0].astype(int)  # labels = [class xywh]
    weights = np.bincount(classes, minlength=nc)  # occurrences per class

    # Prepend gridpoint count (for uCE training)
    # gpi = ((320 / 32 * np.array([1, 2, 4])) ** 2 * 3).sum()  # gridpoints per image
    # weights = np.hstack([gpi * len(labels)  - weights.sum() * 9, weights * 9]) ** 0.5  # prepend gridpoints to start

    weights[weights == 0] = 1  # replace empty bins with 1
    weights = 1 / weights  # number of targets per class
    weights /= weights.sum()  # normalize
    return torch.from_numpy(weights).float()


def labels_to_image_weights(labels, nc=80, class_weights=np.ones(80)):
    # Produces image weights based on class_weights and image contents
    # Usage: index = random.choices(range(n), weights=image_weights, k=1)  # weighted image sample
    class_counts = np.array([np.bincount(x[:, 0].astype(int), minlength=nc) for x in labels])
    return (class_weights.reshape(1, nc) * class_counts).sum(1)


def coco80_to_coco91_class():  # converts 80-index (val2014) to 91-index (paper)
    # https://tech.amikelive.com/node-718/what-object-categories-labels-are-in-coco-dataset/
    # a = np.loadtxt('data/coco.names', dtype='str', delimiter='\n')
    # b = np.loadtxt('data/coco_paper.names', dtype='str', delimiter='\n')
    # x1 = [list(a[i] == b).index(True) + 1 for i in range(80)]  # darknet to coco
    # x2 = [list(b[i] == a).index(True) if any(b[i] == a) else None for i in range(91)]  # coco to darknet
    return [
        1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 27, 28, 31, 32, 33, 34,
        35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
        64, 65, 67, 70, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 84, 85, 86, 87, 88, 89, 90]


def xyxy2xywh(x):
    # Convert nx4 boxes from [x1, y1, x2, y2] to [x, y, w, h] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = (x[:, 0] + x[:, 2]) / 2  # x center
    y[:, 1] = (x[:, 1] + x[:, 3]) / 2  # y center
    y[:, 2] = x[:, 2] - x[:, 0]  # width
    y[:, 3] = x[:, 3] - x[:, 1]  # height
    return y


def xywh2xyxy(x):
    # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
    y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
    y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
    y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y
    return y


def xywhn2xyxy(x, w=640, h=640, padw=0, padh=0):
    # Convert nx4 boxes from [x, y, w, h] normalized to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = w * (x[:, 0] - x[:, 2] / 2) + padw  # top left x
    y[:, 1] = h * (x[:, 1] - x[:, 3] / 2) + padh  # top left y
    y[:, 2] = w * (x[:, 0] + x[:, 2] / 2) + padw  # bottom right x
    y[:, 3] = h * (x[:, 1] + x[:, 3] / 2) + padh  # bottom right y
    return y


def xyxy2xywhn(x, w=640, h=640, clip=False, eps=0.0):
    # Convert nx4 boxes from [x1, y1, x2, y2] to [x, y, w, h] normalized where xy1=top-left, xy2=bottom-right
    if clip:
        clip_boxes(x, (h - eps, w - eps))  # warning: inplace clip
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = ((x[:, 0] + x[:, 2]) / 2) / w  # x center
    y[:, 1] = ((x[:, 1] + x[:, 3]) / 2) / h  # y center
    y[:, 2] = (x[:, 2] - x[:, 0]) / w  # width
    y[:, 3] = (x[:, 3] - x[:, 1]) / h  # height
    return y


def xyn2xy(x, w=640, h=640, padw=0, padh=0):
    # Convert normalized segments into pixel segments, shape (n,2)
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = w * x[:, 0] + padw  # top left x
    y[:, 1] = h * x[:, 1] + padh  # top left y
    return y


def segment2box(segment, width=640, height=640):
    # Convert 1 segment label to 1 box label, applying inside-image constraint, i.e. (xy1, xy2, ...) to (xyxy)
    x, y = segment.T  # segment xy
    inside = (x >= 0) & (y >= 0) & (x <= width) & (y <= height)
    x, y, = x[inside], y[inside]
    return np.array([x.min(), y.min(), x.max(), y.max()]) if any(x) else np.zeros((1, 4))  # xyxy


def segments2boxes(segments):
    # Convert segment labels to box labels, i.e. (cls, xy1, xy2, ...) to (cls, xywh)
    boxes = []
    for s in segments:
        x, y = s.T  # segment xy
        boxes.append([x.min(), y.min(), x.max(), y.max()])  # cls, xyxy
    return xyxy2xywh(np.array(boxes))  # cls, xywh


def resample_segments(segments, n=1000):
    # Up-sample an (n,2) segment
    for i, s in enumerate(segments):
        s = np.concatenate((s, s[0:1, :]), axis=0)
        x = np.linspace(0, len(s) - 1, n)
        xp = np.arange(len(s))
        segments[i] = np.concatenate([np.interp(x, xp, s[:, i]) for i in range(2)]).reshape(2, -1).T  # segment xy
    return segments


def scale_boxes(img1_shape, boxes, img0_shape, ratio_pad=None):
    # Rescale boxes (xyxy) from img1_shape to img0_shape
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    boxes[:, [0, 2]] -= pad[0]  # x padding
    boxes[:, [1, 3]] -= pad[1]  # y padding
    boxes[:, :4] /= gain
    clip_boxes(boxes, img0_shape)
    return boxes


def scale_segments(img1_shape, segments, img0_shape, ratio_pad=None):
    # Rescale coords (xyxy) from img1_shape to img0_shape
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    segments[:, 0] -= pad[0]  # x padding
    segments[:, 1] -= pad[1]  # y padding
    segments /= gain
    clip_segments(segments, img0_shape)
    return segments


def clip_boxes(boxes, shape):
    # Clip boxes (xyxy) to image shape (height, width)
    if isinstance(boxes, torch.Tensor):  # faster individually
        boxes[:, 0].clamp_(0, shape[1])  # x1
        boxes[:, 1].clamp_(0, shape[0])  # y1
        boxes[:, 2].clamp_(0, shape[1])  # x2
        boxes[:, 3].clamp_(0, shape[0])  # y2
    else:  # np.array (faster grouped)
        boxes[:, [0, 2]] = boxes[:, [0, 2]].clip(0, shape[1])  # x1, x2
        boxes[:, [1, 3]] = boxes[:, [1, 3]].clip(0, shape[0])  # y1, y2


def clip_segments(boxes, shape):
    # Clip segments (xy1,xy2,...) to image shape (height, width)
    if isinstance(boxes, torch.Tensor):  # faster individually
        boxes[:, 0].clamp_(0, shape[1])  # x
        boxes[:, 1].clamp_(0, shape[0])  # y
    else:  # np.array (faster grouped)
        boxes[:, 0] = boxes[:, 0].clip(0, shape[1])  # x
        boxes[:, 1] = boxes[:, 1].clip(0, shape[0])  # y


def non_max_suppression(
        prediction,
        conf_thres=0.25,
        iou_thres=0.45,
        classes=None,
        agnostic=False,
        multi_label=False,
        labels=(),
        max_det=300,
        nm=0,  # number of masks
):
    """Non-Maximum Suppression (NMS) on inference results to reject overlapping detections

    Returns:
         list of detections, on (n,6) tensor per image [xyxy, conf, cls]
    """

    if isinstance(prediction, (list, tuple)):  # YOLOv5 model in validation model, output = (inference_out, loss_out)
        prediction = prediction[0]  # select only inference output

    device = prediction.device
    mps = 'mps' in device.type  # Apple MPS
    if mps:  # MPS not fully supported yet, convert tensors to CPU before NMS
        prediction = prediction.cpu()
    bs = prediction.shape[0]  # batch size
    nc = prediction.shape[2] - nm - 5  # number of classes
    xc = prediction[..., 4] > conf_thres  # candidates

    # Checks
    assert 0 <= conf_thres <= 1, f'Invalid Confidence threshold {conf_thres}, valid values are between 0.0 and 1.0'
    assert 0 <= iou_thres <= 1, f'Invalid IoU {iou_thres}, valid values are between 0.0 and 1.0'

    # Settings
    # min_wh = 2  # (pixels) minimum box width and height
    max_wh = 7680  # (pixels) maximum box width and height
    max_nms = 30000  # maximum number of boxes into torchvision.ops.nms()
    time_limit = 0.5 + 0.05 * bs  # seconds to quit after
    redundant = True  # require redundant detections
    multi_label &= nc > 1  # multiple labels per box (adds 0.5ms/img)
    merge = False  # use merge-NMS

    t = time.time()
    mi = 5 + nc  # mask start index
    output = [torch.zeros((0, 6 + nm), device=prediction.device)] * bs
    for xi, x in enumerate(prediction):  # image index, image inference
        # Apply constraints
        # x[((x[..., 2:4] < min_wh) | (x[..., 2:4] > max_wh)).any(1), 4] = 0  # width-height
        x = x[xc[xi]]  # confidence

        # Cat apriori labels if autolabelling
        if labels and len(labels[xi]):
            lb = labels[xi]
            v = torch.zeros((len(lb), nc + nm + 5), device=x.device)
            v[:, :4] = lb[:, 1:5]  # box
            v[:, 4] = 1.0  # conf
            v[range(len(lb)), lb[:, 0].long() + 5] = 1.0  # cls
            x = torch.cat((x, v), 0)

        # If none remain process next image
        if not x.shape[0]:
            continue

        # Compute conf
        x[:, 5:] *= x[:, 4:5]  # conf = obj_conf * cls_conf

        # Box/Mask
        box = xywh2xyxy(x[:, :4])  # center_x, center_y, width, height) to (x1, y1, x2, y2)
        mask = x[:, mi:]  # zero columns if no masks

        # Detections matrix nx6 (xyxy, conf, cls)
        if multi_label:
            i, j = (x[:, 5:mi] > conf_thres).nonzero(as_tuple=False).T
            x = torch.cat((box[i], x[i, 5 + j, None], j[:, None].float(), mask[i]), 1)
        else:  # best class only
            conf, j = x[:, 5:mi].max(1, keepdim=True)
            x = torch.cat((box, conf, j.float(), mask), 1)[conf.view(-1) > conf_thres]

        # Filter by class
        if classes is not None:
            x = x[(x[:, 5:6] == torch.tensor(classes, device=x.device)).any(1)]

        # Apply finite constraint
        # if not torch.isfinite(x).all():
        #     x = x[torch.isfinite(x).all(1)]

        # Check shape
        n = x.shape[0]  # number of boxes
        if not n:  # no boxes
            continue
        elif n > max_nms:  # excess boxes
            x = x[x[:, 4].argsort(descending=True)[:max_nms]]  # sort by confidence
        else:
            x = x[x[:, 4].argsort(descending=True)]  # sort by confidence

        # Batched NMS
        c = x[:, 5:6] * (0 if agnostic else max_wh)  # classes
        boxes, scores = x[:, :4] + c, x[:, 4]  # boxes (offset by class), scores
        i = torchvision.ops.nms(boxes, scores, iou_thres)  # NMS
        if i.shape[0] > max_det:  # limit detections
            i = i[:max_det]
        if merge and (1 < n < 3E3):  # Merge NMS (boxes merged using weighted mean)
            # update boxes as boxes(i,4) = weights(i,n) * boxes(n,4)
            iou = box_iou(boxes[i], boxes) > iou_thres  # iou matrix
            weights = iou * scores[None]  # box weights
            x[i, :4] = torch.mm(weights, x[:, :4]).float() / weights.sum(1, keepdim=True)  # merged boxes
            if redundant:
                i = i[iou.sum(1) > 1]  # require redundancy

        output[xi] = x[i]
        if mps:
            output[xi] = output[xi].to(device)
        if (time.time() - t) > time_limit:
            LOGGER.warning(f'WARNING ⚠️ NMS time limit {time_limit:.3f}s exceeded')
            break  # time limit exceeded

    return output


def strip_optimizer(f='best.pt', s=''):  # from utils.general import *; strip_optimizer()
    # Strip optimizer from 'f' to finalize training, optionally save as 's'
    x = torch.load(f, map_location=torch.device('cpu'))
    if x.get('ema'):
        x['model'] = x['ema']  # replace model with ema
    for k in 'optimizer', 'best_fitness', 'wandb_id', 'ema', 'updates':  # keys
        x[k] = None
    x['epoch'] = -1
    x['model'].half()  # to FP16
    for p in x['model'].parameters():
        p.requires_grad = False
    torch.save(x, s or f)
    mb = os.path.getsize(s or f) / 1E6  # filesize
    LOGGER.info(f"Optimizer stripped from {f},{f' saved as {s},' if s else ''} {mb:.1f}MB")


def print_mutation(keys, results, hyp, save_dir, bucket, prefix=colorstr('evolve: ')):
    evolve_csv = save_dir / 'evolve.csv'
    evolve_yaml = save_dir / 'hyp_evolve.yaml'
    keys = tuple(keys) + tuple(hyp.keys())  # [results + hyps]
    keys = tuple(x.strip() for x in keys)
    vals = results + tuple(hyp.values())
    n = len(keys)

    # Download (optional)
    if bucket:
        url = f'gs://{bucket}/evolve.csv'
        if gsutil_getsize(url) > (evolve_csv.stat().st_size if evolve_csv.exists() else 0):
            os.system(f'gsutil cp {url} {save_dir}')  # download evolve.csv if larger than local

    # Log to evolve.csv
    s = '' if evolve_csv.exists() else (('%20s,' * n % keys).rstrip(',') + '\n')  # add header
    with open(evolve_csv, 'a') as f:
        f.write(s + ('%20.5g,' * n % vals).rstrip(',') + '\n')

    # Save yaml
    with open(evolve_yaml, 'w') as f:
        data = pd.read_csv(evolve_csv)
        data = data.rename(columns=lambda x: x.strip())  # strip keys
        i = np.argmax(fitness(data.values[:, :4]))  #
        generations = len(data)
        f.write('# YOLOv5 Hyperparameter Evolution Results\n' + f'# Best generation: {i}\n' +
                f'# Last generation: {generations - 1}\n' + '# ' + ', '.join(f'{x.strip():>20s}' for x in keys[:7]) +
                '\n' + '# ' + ', '.join(f'{x:>20.5g}' for x in data.values[i, :7]) + '\n\n')
        yaml.safe_dump(data.loc[i][7:].to_dict(), f, sort_keys=False)

    # Print to screen
    LOGGER.info(prefix + f'{generations} generations finished, current result:\n' + prefix +
                ', '.join(f'{x.strip():>20s}' for x in keys) + '\n' + prefix + ', '.join(f'{x:20.5g}'
                                                                                         for x in vals) + '\n\n')

    if bucket:
        os.system(f'gsutil cp {evolve_csv} {evolve_yaml} gs://{bucket}')  # upload


def apply_classifier(x, model, img, im0):
    # Apply a second stage classifier to YOLO outputs
    # Example model = torchvision.models.__dict__['efficientnet_b0'](pretrained=True).to(device).eval()
    im0 = [im0] if isinstance(im0, np.ndarray) else im0
    for i, d in enumerate(x):  # per image
        if d is not None and len(d):
            d = d.clone()

            # Reshape and pad cutouts
            b = xyxy2xywh(d[:, :4])  # boxes
            b[:, 2:] = b[:, 2:].max(1)[0].unsqueeze(1)  # rectangle to square
            b[:, 2:] = b[:, 2:] * 1.3 + 30  # pad
            d[:, :4] = xywh2xyxy(b).long()

            # Rescale boxes from img_size to im0 size
            scale_boxes(img.shape[2:], d[:, :4], im0[i].shape)

            # Classes
            pred_cls1 = d[:, 5].long()
            ims = []
            for a in d:
                cutout = im0[i][int(a[1]):int(a[3]), int(a[0]):int(a[2])]
                im = cv2.resize(cutout, (224, 224))  # BGR

                im = im[:, :, ::-1].transpose(2, 0, 1)  # BGR to RGB, to 3x416x416
                im = np.ascontiguousarray(im, dtype=np.float32)  # uint8 to float32
                im /= 255  # 0 - 255 to 0.0 - 1.0
                ims.append(im)

            pred_cls2 = model(torch.Tensor(ims).to(d.device)).argmax(1)  # classifier prediction
            x[i] = x[i][pred_cls1 == pred_cls2]  # retain matching class detections

    return x


def increment_path(path, exist_ok=False, sep='', mkdir=False):
    # Increment file or directory path, i.e. runs/exp --> runs/exp{sep}2, runs/exp{sep}3, ... etc.
    path = Path(path)  # os-agnostic
    if path.exists() and not exist_ok:
        path, suffix = (path.with_suffix(''), path.suffix) if path.is_file() else (path, '')

        # Method 1
        for n in range(2, 9999):
            p = f'{path}{sep}{n}{suffix}'  # increment path
            if not os.path.exists(p):  #
                break
        path = Path(p)

        # Method 2 (deprecated)
        # dirs = glob.glob(f"{path}{sep}*")  # similar paths
        # matches = [re.search(rf"{path.stem}{sep}(\d+)", d) for d in dirs]
        # i = [int(m.groups()[0]) for m in matches if m]  # indices
        # n = max(i) + 1 if i else 2  # increment number
        # path = Path(f"{path}{sep}{n}{suffix}")  # increment path

    if mkdir:
        path.mkdir(parents=True, exist_ok=True)  # make directory

    return path


# OpenCV Chinese-friendly functions ------------------------------------------------------------------------------------
imshow_ = cv2.imshow  # copy to avoid recursion errors


def imread(path, flags=cv2.IMREAD_COLOR):
    return cv2.imdecode(np.fromfile(path, np.uint8), flags)


def imwrite(path, im):
    try:
        cv2.imencode(Path(path).suffix, im)[1].tofile(path)
        return True
    except Exception:
        return False


def imshow(path, im):
    imshow_(path.encode('unicode_escape').decode(), im)


cv2.imread, cv2.imwrite, cv2.imshow = imread, imwrite, imshow  # redefine

# Variables ------------------------------------------------------------------------------------------------------------
NCOLS = 0 if is_docker() else shutil.get_terminal_size().columns  # terminal window size for tqdm



from .decode_head import decode_yolo_head
try:
    from .inference import InferenceSession
    from .device import KneronDevice
except:
    print("[warning] kp is not install")

try:
    from .convert import onnx_to_nef, combine_nef_files
except ImportError:
    print("[Warning] Kneron toolchain not found. Please install it when want to use the convert function.")